"""Sources."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from tox.config.loader.section import Section

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from tox.config.loader.api import Loader, OverrideMap
    from tox.config.sets import ConfigSet, CoreConfigSet


class Source(ABC):
    """Source is able to return a configuration value (for either the core or per environment source)."""

    FILENAME = ""

    def __init__(self, path: Path) -> None:
        self.path: Path = path  #: the path to the configuration source
        self._section_to_loaders: dict[str, list[Loader[Any]]] = {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path})"

    def get_loaders(
        self,
        section: Section,
        base: list[str] | None,
        override_map: OverrideMap,
        conf: ConfigSet,
    ) -> Iterator[Loader[Any]]:
        """Return a loader that loads settings from a given section name.

        :param section: the section to load
        :param base: base sections to fallback to
        :param override_map: a list of overrides to apply
        :param conf: the config set to use

        :returns: the loaders to use

        """
        section = self.transform_section(section)
        key = section.key
        if key in self._section_to_loaders:
            yield from self._section_to_loaders[key]
            return
        loaders: list[Loader[Any]] = []
        self._section_to_loaders[key] = loaders
        loader: Loader[Any] | None = self.get_loader(section, override_map)
        if loader is not None:
            loaders.append(loader)
            yield loader

        if base is not None:
            conf.add_config(
                keys="base",
                of_type=list[str],
                desc="inherit missing keys from these sections",
                default=base,
            )
            for base_section in self.get_base_sections(conf["base"], section):
                child = loader
                loader = self.get_loader(base_section, override_map)
                if loader is None:
                    loader = child
                    continue
                if child is not None and loader is not None:
                    child.parent = loader
                yield loader
                loaders.append(loader)

    @abstractmethod
    def transform_section(self, section: Section) -> Section:
        raise NotImplementedError

    @abstractmethod
    def get_loader(self, section: Section, override_map: OverrideMap) -> Loader[Any] | None:
        raise NotImplementedError

    @abstractmethod
    def get_base_sections(self, base: list[str], in_section: Section) -> Iterator[Section]:
        raise NotImplementedError

    @abstractmethod
    def sections(self) -> Iterator[Section]:
        """Return a loader that loads the core configuration values.

        :returns: the core loader from this source

        """
        raise NotImplementedError

    @abstractmethod
    def envs(self, core_conf: CoreConfigSet) -> Iterator[str]:
        """:param core_conf: the core configuration set

        :returns: a list of environments defined within this source

        """
        raise NotImplementedError

    @abstractmethod
    def get_tox_env_section(self, item: str) -> tuple[Section, list[str], list[str]]:
        """:returns: the section for a tox environment"""
        raise NotImplementedError

    @abstractmethod
    def get_core_section(self) -> Section:
        """:returns: the core section"""
        raise NotImplementedError


__all__ = [
    "Section",
    "Source",
]



from __future__ import annotations

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

from tox.config.set_env import SetEnv
from tox.config.types import Command, EnvList
from tox.tox_env.python.pip.req_file import PythonDeps


def to_native(value: Any) -> Any:
    if isinstance(value, bool):
        return value
    if isinstance(value, (int, float, str)):
        return value
    if isinstance(value, Path):
        return str(value)
    return _to_native_complex(value)


def _to_native_complex(value: Any) -> Any:
    if isinstance(value, SetEnv):
        return {k: to_native(value.load(k)) for k in sorted(value)}
    if isinstance(value, PythonDeps):
        return to_native(value.lines())
    if isinstance(value, EnvList):
        return value.envs
    if isinstance(value, Command):
        return value.shell
    return _to_native_collection(value)


def _to_native_collection(value: Any) -> Any:
    if isinstance(value, Mapping):
        return {str(k): to_native(v) for k, v in value.items()}
    if isinstance(value, set):
        return sorted(str(i) for i in value)
    if isinstance(value, Sequence):
        return [to_native(i) for i in value]
    return str(value)


__all__ = ("to_native",)



from __future__ import annotations

import contextlib
import errno
import json
import locale
import logging
import os
import re
import selectors
import shutil
import stat
import subprocess
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import TYPE_CHECKING, Any, NoReturn
from unittest.mock import MagicMock, create_autospec

import psutil
import pytest
from colorama import Fore
from psutil import AccessDenied

from tox.execute.api import ExecuteOptions, Outcome
from tox.execute.local_sub_process import SIG_INTERRUPT, LocalSubProcessExecuteInstance, LocalSubProcessExecutor
from tox.execute.local_sub_process.read_via_thread_unix import ReadViaThreadUnix
from tox.execute.request import ExecuteRequest, StdinSource
from tox.execute.stream import SyncWrite
from tox.report import NamedBytesIO

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tox.pytest import CaptureFixture, LogCaptureFixture, MonkeyPatch


class FakeOutErr:
    def __init__(self) -> None:
        self.out_err = (
            TextIOWrapper(NamedBytesIO("out"), encoding=locale.getpreferredencoding(False)),
            TextIOWrapper(NamedBytesIO("err"), encoding=locale.getpreferredencoding(False)),
        )

    def read_out_err(self) -> tuple[str, str]:
        out_got = self.out_err[0].buffer.getvalue().decode(self.out_err[0].encoding)
        err_got = self.out_err[1].buffer.getvalue().decode(self.out_err[1].encoding)
        return out_got, err_got


def _create_mock_env() -> MagicMock:
    """Create a mock tox environment with no_capture=False to prevent console inheritance."""
    mock_env = MagicMock()
    mock_env.options.no_capture = False
    return mock_env


@pytest.mark.parametrize("color", [True, False], ids=["color", "no_color"])
@pytest.mark.parametrize(("out", "err"), [("out", "err"), ("", "")], ids=["simple", "nothing"])
@pytest.mark.parametrize("show", [True, False], ids=["show", "no_show"])
@pytest.mark.parametrize(
    "stderr_color",
    ["RED", "YELLOW", "RESET"],
    ids=["stderr_color_default", "stderr_color_yellow", "stderr_color_reset"],
)
def test_local_execute_basic_pass(  # noqa: PLR0913
    caplog: LogCaptureFixture,
    os_env: dict[str, str],
    out: str,
    err: str,
    show: bool,
    color: bool,
    stderr_color: str,
) -> None:
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=color)

    tox_env = _create_mock_env()
    tox_env.conf._conf.options.stderr_color = stderr_color  # noqa: SLF001
    code = f"import sys; print({out!r}, end=''); print({err!r}, end='', file=sys.stderr)"
    request = ExecuteRequest(cmd=[sys.executable, "-c", code], cwd=Path(), env=os_env, stdin=StdinSource.OFF, run_id="")
    out_err = FakeOutErr()

    with executor.call(request, show=show, out_err=out_err.out_err, env=tox_env) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()
    assert status.out == out.encode()
    assert status.err == err.encode()
    outcome = status.outcome
    assert outcome is not None
    assert bool(outcome) is True, outcome
    assert outcome.exit_code == Outcome.OK
    assert outcome.err == err
    assert outcome.out == out
    assert outcome.request == request

    out_got, err_got = out_err.read_out_err()
    if show:
        assert out_got == out
        expected = f"{getattr(Fore, stderr_color)}{err}{Fore.RESET}" if color and err else err
        assert err_got == expected
    else:
        assert not out_got
        assert not err_got
    assert not caplog.records


def test_local_execute_basic_pass_show_on_standard_newline_flush(caplog: LogCaptureFixture) -> None:
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=False)
    request = ExecuteRequest(
        cmd=[sys.executable, "-c", "import sys; print('out'); print('yay')"],
        cwd=Path(),
        env=os.environ.copy(),
        stdin=StdinSource.OFF,
        run_id="",
    )
    out_err = FakeOutErr()
    with executor.call(request, show=True, out_err=out_err.out_err, env=_create_mock_env()) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()
    outcome = status.outcome
    assert outcome is not None
    assert repr(outcome)
    assert bool(outcome) is True, outcome
    assert outcome.exit_code == Outcome.OK
    assert not outcome.err
    assert outcome.out == f"out{os.linesep}yay{os.linesep}"
    out, err = out_err.read_out_err()
    assert out == f"out{os.linesep}yay{os.linesep}"
    assert not err
    assert not caplog.records


def test_local_execute_write_a_lot(os_env: dict[str, str]) -> None:
    count = 10_000
    executor = LocalSubProcessExecutor(colored=False)
    request = ExecuteRequest(
        cmd=[
            sys.executable,
            "-c",
            (
                "import sys; import time; from datetime import datetime; import os;"
                f"print('e' * {count}, file=sys.stderr);"
                f"print('o' * {count}, file=sys.stdout);"
                "time.sleep(0.5);"
                f"print('a' * {count}, file=sys.stderr);"
                f"print('b' * {count}, file=sys.stdout);"
            ),
        ],
        cwd=Path(),
        env=os_env,
        stdin=StdinSource.OFF,
        run_id="",
    )
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err, env=_create_mock_env()) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()
    outcome = status.outcome
    assert outcome is not None
    assert bool(outcome), outcome
    expected_out = f"{'o' * count}{os.linesep}{'b' * count}{os.linesep}"
    assert outcome.out == expected_out, expected_out[len(outcome.out) :]
    expected_err = f"{'e' * count}{os.linesep}{'a' * count}{os.linesep}"
    assert outcome.err == expected_err, expected_err[len(outcome.err) :]


@pytest.mark.skipif(sys.platform == "win32", reason="Unix terminal size test")
def test_local_execute_terminal_size(os_env: dict[str, str], monkeypatch: MonkeyPatch) -> None:
    """Regression test for #2999 - check terminal size is set correctly in tox subprocess."""
    import pty  # noqa: PLC0415

    terminal_size = os.terminal_size((84, 42))
    main, child = pty.openpty()  # ty: ignore[possibly-missing-attribute] # Unix-only

    # Use ReadViaThreadUnix to help with debugging the test itself.
    pipe_out = ReadViaThreadUnix(main, sys.stdout.buffer.write, name="testout", drain=True)
    with (
        pipe_out,
        monkeypatch.context() as monkey,
        open(  # noqa: PTH123
            child, "w", encoding=locale.getpreferredencoding(False)
        ) as stdout_mock,
    ):
        # Switch stdout with test pty
        monkey.setattr(sys, "stdout", stdout_mock)
        monkey.setenv("COLUMNS", "84")
        monkey.setenv("LINES", "42")

        executor = LocalSubProcessExecutor(colored=False)
        request = ExecuteRequest(
            cmd=[sys.executable, "-c", "import os; print(os.get_terminal_size())"],
            cwd=Path(),
            env=os_env,
            stdin=StdinSource.OFF,
            run_id="",
        )
        out_err = FakeOutErr()
        with executor.call(request, show=False, out_err=out_err.out_err, env=_create_mock_env()) as status:
            while status.exit_code is None:  # pragma: no branch
                status.wait()
    outcome = status.outcome
    assert outcome is not None
    assert bool(outcome), outcome
    expected_out = f"{terminal_size!r}\r\n"
    assert outcome.out == expected_out, expected_out[len(outcome.out) :]
    assert not outcome.err


def test_local_execute_basic_fail(capsys: CaptureFixture, caplog: LogCaptureFixture, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.chdir(Path(__file__).parents[3])
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=False)
    cwd = Path().absolute()
    cmd = [
        sys.executable,
        "-c",
        "import sys; print('out', end=''); print('err', file=sys.stderr, end=''); sys.exit(3)",
    ]
    request = ExecuteRequest(cmd=cmd, cwd=cwd, env=os.environ.copy(), stdin=StdinSource.OFF, run_id="")

    # run test
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err, env=_create_mock_env()) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()
    outcome = status.outcome
    assert outcome is not None

    assert repr(outcome)

    # assert no output, no logs
    out, err = out_err.read_out_err()
    assert not out
    assert not err
    assert not caplog.records

    # assert return object
    assert bool(outcome) is False, outcome
    assert outcome.exit_code == 3
    assert outcome.err == "err"
    assert outcome.out == "out"
    assert outcome.request == request

    # asset fail
    with pytest.raises(SystemExit) as context:
        outcome.assert_success()
    # asset fail
    assert context.value.code == 3

    out, err = capsys.readouterr()
    assert out == "out\n"
    expected = f"{Fore.RED}err{Fore.RESET}\n"
    assert err == expected

    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.CRITICAL
    assert record.msg == "exit %s (%.2f seconds) %s> %s%s"
    assert record.args is not None
    code, duration, cwd_, cmd_, metadata = record.args
    assert code == 3
    assert cwd_ == cwd
    assert cmd_ == request.shell_cmd
    assert isinstance(duration, float)
    assert duration > 0
    assert isinstance(metadata, str)
    assert metadata.startswith(" pid=")


def test_command_does_not_exist(caplog: LogCaptureFixture, os_env: dict[str, str]) -> None:
    caplog.set_level(logging.NOTSET)
    executor = LocalSubProcessExecutor(colored=False)
    request = ExecuteRequest(
        cmd=["sys-must-be-missing"],
        cwd=Path().absolute(),
        env=os_env,
        stdin=StdinSource.OFF,
        run_id="",
    )
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err, env=_create_mock_env()) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()  # pragma: no cover
    outcome = status.outcome
    assert outcome is not None

    assert bool(outcome) is False, outcome
    assert outcome.exit_code != Outcome.OK
    assert not outcome.out
    assert not outcome.err
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == "ERROR"
    assert re.match(
        r".*(No such file or directory|The system cannot find the file specified).*", caplog.records[0].message
    )


@pytest.mark.skipif(sys.platform == "win32", reason="You need a conhost shell for keyboard interrupt")
@pytest.mark.flaky(max_runs=3, min_passes=1)
def test_command_keyboard_interrupt(tmp_path: Path, monkeypatch: MonkeyPatch, capfd: CaptureFixture) -> None:
    monkeypatch.chdir(tmp_path)
    process_up_signal = tmp_path / "signal"
    cmd = [sys.executable, str(Path(__file__).parent / "local_subprocess_sigint.py"), str(process_up_signal)]
    process = subprocess.Popen(cmd)
    while not process_up_signal.exists():
        assert process.poll() is None
    root = process.pid
    try:
        child = next(iter(psutil.Process(pid=root).children())).pid
    except AccessDenied as exc:  # pragma: no cover # on termux for example
        pytest.skip(str(exc))  # pragma: no cover
        raise  # pragma: no cover

    print(f"test running in {os.getpid()} and sending CTRL+C to {process.pid}", file=sys.stderr)  # noqa: T201
    process.send_signal(SIG_INTERRUPT)
    try:
        process.communicate(timeout=3)
    except subprocess.TimeoutExpired:  # pragma: no cover
        process.kill()
        raise

    out, err = capfd.readouterr()
    assert f"W	requested interrupt of {child} from {root}, activate in 0.01" in err, err
    assert f"W	send signal SIGINT(2) to {child} from {root} with timeout 0.05" in err, err
    assert f"W	send signal SIGTERM(15) to {child} from {root} with timeout 0.07" in err, err
    assert f"W	send signal SIGKILL(9) to {child} from {root}" in err, err

    outs = out.split("\n")

    exit_code = int(outs[0])
    assert exit_code == -9
    assert float(outs[3]) > 0  # duration
    assert "how about no signal 2" in outs[1], outs[1]  # 2 - Interrupt
    assert "how about no signal 15" in outs[1], outs[1]  # 15 - Terminated


@pytest.mark.parametrize("tty_mode", ["on", "off"])
def test_local_subprocess_tty(monkeypatch: MonkeyPatch, mocker: MockerFixture, tty_mode: str) -> None:
    monkeypatch.setenv("COLUMNS", "100")
    monkeypatch.setenv("LINES", "100")
    tty = tty_mode == "on"
    mocker.patch("sys.stdout.isatty", return_value=tty)
    mocker.patch("sys.stderr.isatty", return_value=tty)
    try:
        import termios  # noqa: F401, PLC0415
    except ImportError:
        exp_tty = False  # platforms without tty support at all
    else:
        # to avoid trying (and failing) to copy mode bits
        exp_tty = tty
        mocker.patch("termios.tcgetattr")
        mocker.patch("termios.tcsetattr")

    executor = LocalSubProcessExecutor(colored=False)
    cmd: list[str] = [sys.executable, str(Path(__file__).parent / "tty_check.py")]
    request = ExecuteRequest(cmd=cmd, stdin=StdinSource.API, cwd=Path.cwd(), env=dict(os.environ), run_id="")
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err, env=_create_mock_env()) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()
    outcome = status.outcome
    assert outcome is not None

    assert outcome
    info = json.loads(outcome.out)
    assert info == {
        "stdout": exp_tty,
        "stderr": exp_tty,
        "stdin": False,
        "terminal": [100, 100],
    }


@pytest.mark.parametrize("mode", ["stem", "full", "stem-pattern", "full-pattern", "all"])
def test_allow_list_external_ok(fake_exe_on_path: Path, mode: str) -> None:
    exe = f"{fake_exe_on_path}{'.EXE' if sys.platform == 'win32' else ''}"
    allow = exe if "full" in mode else fake_exe_on_path.stem
    allow = f"{allow[:-2]}*" if "pattern" in mode else allow
    allow = "*" if mode == "all" else allow

    request = ExecuteRequest(
        cmd=[fake_exe_on_path.stem],
        cwd=Path.cwd(),
        env={"PATH": os.environ["PATH"]},
        stdin=StdinSource.OFF,
        run_id="run-id",
        allow=[allow],
    )
    inst = LocalSubProcessExecuteInstance(request, MagicMock(), out=SyncWrite("out", None), err=SyncWrite("err", None))

    assert inst.cmd == [exe]


def test_shebang_limited_on(tmp_path: Path) -> None:
    exe, script, instance = _create_shebang_test(tmp_path, env={"TOX_LIMITED_SHEBANG": "1"})
    if sys.platform == "win32":  # pragma: win32 cover
        assert instance.cmd == [str(script), "--magic"]
    else:
        assert instance.cmd == [exe, "-s", str(script), "--magic"]


@pytest.mark.parametrize("env", [{}, {"TOX_LIMITED_SHEBANG": ""}])
def test_shebang_limited_off(tmp_path: Path, env: dict[str, str]) -> None:
    _, script, instance = _create_shebang_test(tmp_path, env=env)
    assert instance.cmd == [str(script), "--magic"]


def test_shebang_failed_to_parse(tmp_path: Path) -> None:
    _, script, instance = _create_shebang_test(tmp_path, env={"TOX_LIMITED_SHEBANG": "yes"})
    script.write_text("")
    assert instance.cmd == [str(script), "--magic"]


def _create_shebang_test(tmp_path: Path, env: dict[str, str]) -> tuple[str, Path, LocalSubProcessExecuteInstance]:
    exe = shutil.which("python3") or shutil.which("python")
    assert exe is not None
    script = tmp_path / f"s{'.EXE' if sys.platform == 'win32' else ''}"
    script.write_text(f"#!{exe} -s")
    script.chmod(script.stat().st_mode | stat.S_IEXEC)  # mark it executable
    env["PATH"] = str(script.parent)
    request = create_autospec(ExecuteRequest, cmd=["s", "--magic"], env=env, allow=None)
    writer = create_autospec(SyncWrite)
    instance = LocalSubProcessExecuteInstance(request, create_autospec(ExecuteOptions), writer, writer)
    return exe, script, instance


@pytest.mark.parametrize("key", ["COLUMNS", "ROWS"])
def test_local_execute_does_not_overwrite(key: str, mocker: MockerFixture) -> None:
    mocker.patch("shutil.get_terminal_size", return_value=(101, 102))
    env = dict(os.environ)
    env[key] = key
    executor = LocalSubProcessExecutor(colored=False)
    cmd = [sys.executable, "-c", f"import os; print(os.environ['{key}'], end='')"]
    request = ExecuteRequest(cmd=cmd, stdin=StdinSource.API, cwd=Path.cwd(), env=env, run_id="")
    out_err = FakeOutErr()
    with executor.call(request, show=False, out_err=out_err.out_err, env=_create_mock_env()) as status:
        while status.exit_code is None:  # pragma: no branch
            status.wait()
    outcome = status.outcome

    assert outcome is not None
    assert outcome.out == key


@pytest.mark.skipif(sys.platform == "win32", reason="Unix-specific tests")
def test_read_via_thread_eintr_during_select(mocker: MockerFixture) -> None:
    """Test that EINTR during selector.select() is handled correctly."""
    # Create a pipe
    read_fd, write_fd = os.pipe()
    try:
        data_received = []
        eintr_raised = [False]

        def handler(data: bytes) -> int:
            data_received.append(data)
            return len(data)

        # Mock selector.select to raise OSError with EINTR once
        original_select = selectors.DefaultSelector.select

        def mock_select(self: Any, timeout: float | None = None) -> list[tuple[selectors.SelectorKey, int]]:
            if not eintr_raised[0]:
                eintr_raised[0] = True
                err = OSError("Interrupted system call")
                err.errno = errno.EINTR
                raise err
            return original_select(self, timeout)

        mocker.patch.object(selectors.DefaultSelector, "select", mock_select)

        # Write some data
        os.write(write_fd, b"test data")
        os.close(write_fd)

        # Read the data
        reader = ReadViaThreadUnix(read_fd, handler, "test", drain=True)
        with reader:
            reader.thread.join(timeout=2)

        assert eintr_raised[0], "EINTR should have been raised"
        assert b"test data" in data_received, "Data should still be read after EINTR"
    finally:
        with contextlib.suppress(OSError):
            os.close(read_fd)


@pytest.mark.skipif(sys.platform == "win32", reason="Unix-specific tests")
def test_read_via_thread_eintr_during_read(mocker: MockerFixture) -> None:
    """Test that EINTR during os.read() is handled correctly."""
    # Create a pipe
    read_fd, write_fd = os.pipe()
    try:
        data_received = []

        def handler(data: bytes) -> int:
            data_received.append(data)
            return len(data)

        # Mock os.read to raise EINTR once, then work normally
        original_read = os.read
        call_count = [0]

        def mock_read(fd: int, n: int) -> bytes:
            call_count[0] += 1
            if call_count[0] == 1:
                err = OSError("Interrupted")
                err.errno = errno.EINTR
                raise err
            return original_read(fd, n)

        mocker.patch("os.read", mock_read)

        # Write some data
        os.write(write_fd, b"test data")
        os.close(write_fd)

        # Read the data
        reader = ReadViaThreadUnix(read_fd, handler, "test", drain=True)
        with reader:
            reader.thread.join(timeout=1)

        assert b"test data" in data_received
    finally:
        with contextlib.suppress(OSError):
            os.close(read_fd)


@pytest.mark.skipif(sys.platform == "win32", reason="Unix-specific tests")
def test_read_via_thread_ebadf_during_read(mocker: MockerFixture) -> None:
    """Test that EBADF during os.read() is handled correctly."""
    # Create a pipe
    read_fd, write_fd = os.pipe()
    try:
        data_received = []

        def handler(data: bytes) -> int:
            data_received.append(data)
            return len(data)

        # Mock os.read to raise EBADF
        def mock_read(fd: int, n: int) -> NoReturn:  # noqa: ARG001
            err = OSError("Bad file descriptor")
            err.errno = errno.EBADF
            raise err

        mocker.patch("os.read", mock_read)

        # Close write end
        os.close(write_fd)

        # Try to read - should handle EBADF gracefully
        reader = ReadViaThreadUnix(read_fd, handler, "test", drain=True)
        with reader:
            reader.thread.join(timeout=1)

        # Should not crash, but won't receive data due to EBADF
    finally:
        with contextlib.suppress(OSError):
            os.close(read_fd)


@pytest.mark.skipif(sys.platform == "win32", reason="Unix-specific tests")
def test_read_via_thread_drain_no_data() -> None:
    """Test drain_stream when there's no data ready."""
    # Create a pipe with no data
    read_fd, write_fd = os.pipe()
    try:
        data_received = []

        def handler(data: bytes) -> int:
            data_received.append(data)
            return len(data)

        # Close write end immediately so no data will be ready
        os.close(write_fd)

        reader = ReadViaThreadUnix(read_fd, handler, "test", drain=True)
        # Call drain directly to test the "if not ready: break" path
        reader._drain_stream()  # noqa: SLF001

        # Should handle gracefully with no data
        assert len(data_received) == 0
    finally:
        with contextlib.suppress(OSError):
            os.close(read_fd)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from tests.config.loader.conftest import ReplaceOne


def test_replace_glob_matches(replace_one: ReplaceOne, tmp_path: Path) -> None:
    (tmp_path / "a.txt").touch()
    (tmp_path / "b.txt").touch()
    result = replace_one(f"{{glob:{tmp_path}/*.txt}}")
    assert str(tmp_path / "a.txt") in result
    assert str(tmp_path / "b.txt") in result


def test_replace_glob_sorted(replace_one: ReplaceOne, tmp_path: Path) -> None:
    (tmp_path / "c.txt").touch()
    (tmp_path / "a.txt").touch()
    (tmp_path / "b.txt").touch()
    result = replace_one(f"{{glob:{tmp_path}/*.txt}}")
    paths = result.split()
    assert paths == sorted(paths)


def test_replace_glob_no_matches_empty(replace_one: ReplaceOne, tmp_path: Path) -> None:
    result = replace_one(f"{{glob:{tmp_path}/*.xyz}}")
    assert not result


def test_replace_glob_no_matches_default(replace_one: ReplaceOne, tmp_path: Path) -> None:
    result = replace_one(f"{{glob:{tmp_path}/*.xyz:fallback}}")
    assert result == "fallback"


def test_replace_glob_relative_path(replace_one: ReplaceOne, tmp_path: Path) -> None:
    (tmp_path / "dist").mkdir()
    (tmp_path / "dist" / "pkg.whl").touch()
    result = replace_one("{glob:dist/*.whl}")
    assert result == str(tmp_path / "dist" / "pkg.whl")


def test_replace_glob_recursive(replace_one: ReplaceOne, tmp_path: Path) -> None:
    sub = tmp_path / "src" / "pkg"
    sub.mkdir(parents=True)
    (sub / "mod.py").touch()
    pattern = tmp_path / "src" / "**" / "*.py"
    result = replace_one(f"{{glob:{pattern}}}")
    assert str(sub / "mod.py") in result


def test_replace_glob_no_pattern_error(replace_one: ReplaceOne) -> None:
    with pytest.raises(Exception, match="No pattern was supplied"):
        replace_one("{glob:}")


def test_replace_glob_nested_substitution(replace_one: ReplaceOne, tmp_path: Path) -> None:
    (tmp_path / "marker.txt").touch()
    result = replace_one("{glob:{tox_root}/*.txt}")
    assert str(tmp_path / "marker.txt") in result



"""A pytest plugin useful to test tox itself (and its plugins)."""

from __future__ import annotations

import inspect
import os
import re
import shutil
import socket
import sys
import textwrap
import warnings
from contextlib import closing, contextmanager
from pathlib import Path
from types import ModuleType, TracebackType
from typing import TYPE_CHECKING, Any, Protocol, cast

import pytest
from _pytest.fixtures import SubRequest  # noqa: PLC2701
from devpi_process import IndexServer
from virtualenv.info import fs_supports_symlink

import tox.run
from tox.execute.api import Execute, ExecuteInstance, ExecuteOptions, ExecuteStatus, Outcome
from tox.execute.request import ExecuteRequest, shell_cmd
from tox.plugin import manager
from tox.report import LOGGER, OutErr
from tox.run import run as tox_run
from tox.run import setup_state as previous_setup_state
from tox.session.cmd.run.parallel import ENV_VAR_KEY
from tox.tox_env import api as tox_env_api
from tox.tox_env.api import ToxEnv

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence
    from unittest.mock import MagicMock

    from pytest_mock import MockerFixture

    from tox.config.sets import EnvConfigSet
    from tox.execute.stream import SyncWrite
    from tox.session.state import State

    CaptureFixture = pytest.CaptureFixture[str]
else:
    CaptureFixture = pytest.CaptureFixture

os.environ["PIP_DISABLE_PIP_VERSION_CHECK"] = "1"
os.environ["PIP_NO_PYTHON_VERSION_WARNING"] = "1"

if fs_supports_symlink():  # pragma: no cover # used to speed up test suite run time where possible
    os.environ["VIRTUALENV_SYMLINK_APP_DATA"] = "1"
    os.environ["VIRTUALENV_SYMLINKS"] = "1"


@pytest.fixture(autouse=True)
def ensure_logging_framework_not_altered() -> Iterator[None]:
    before_handlers = list(LOGGER.handlers)
    yield
    LOGGER.handlers = before_handlers


@pytest.fixture(autouse=True)
def _disable_root_tox_py(request: SubRequest, mocker: MockerFixture) -> Iterator[None]:
    """Unless this is a plugin test do not allow loading toxfile.py."""
    if request.node.get_closest_marker("plugin_test"):  # unregister inline plugin
        module, load_inline = None, manager._load_inline  # noqa: SLF001

        def _load_inline(path: Path) -> ModuleType | None:  # register only on first run, and unregister at end
            nonlocal module
            module = load_inline(path)
            return module

        mocker.patch.object(manager, "_load_inline", _load_inline)
        yield
        if module is not None and manager.MANAGER.manager.is_registered(module):  # pragma: no branch
            manager.MANAGER.manager.unregister(module)
    else:  # do not allow loading inline plugins
        mocker.patch("tox.plugin.inline._load_plugin", return_value=None)
        yield


@contextmanager
def check_os_environ() -> Iterator[None]:
    old = os.environ.copy()
    to_clean = {
        k: os.environ.pop(k, None)
        for k in (ENV_VAR_KEY, "TOX_WORK_DIR", "PYTHONPATH", "COV_CORE_CONTEXT", "TOX_DISABLED_EXTERNAL_PLUGINS")
    }

    yield

    for key, value in to_clean.items():
        if value is not None:
            os.environ[key] = value

    new = os.environ
    extra = {k: new[k] for k in set(new) - set(old)}
    extra.pop("PLAT", None)
    extra.pop("TOX_DISABLED_EXTERNAL_PLUGINS", None)
    miss = {k: old[k] for k in set(old) - set(new)}
    diff = {
        f"{k} = {old[k]} vs {new[k]}" for k in set(old) & set(new) if old[k] != new[k] and not k.startswith("PYTEST_")
    }
    if extra or miss or diff:
        msg = "test changed environ"
        if extra:
            msg += f" extra {extra}"
        if miss:
            msg += f" miss {miss}"
        if diff:
            msg += f" diff {diff}"
        pytest.fail(msg)


@pytest.fixture(autouse=True)
def check_os_environ_stable(monkeypatch: pytest.MonkeyPatch) -> Iterator[None]:
    with check_os_environ():
        yield
        monkeypatch.undo()


@pytest.fixture(autouse=True)
def no_color(monkeypatch: pytest.MonkeyPatch, check_os_environ_stable: None) -> None:  # noqa: ARG001
    monkeypatch.setenv("NO_COLOR", "yes")


class ToxProject:
    def __init__(  # noqa: PLR0913
        self,
        files: dict[str, Any],
        base: Path | None,
        path: Path,
        capfd: CaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
        mocker: MockerFixture,
    ) -> None:
        self.path: Path = path
        self.monkeypatch: pytest.MonkeyPatch = monkeypatch
        self.mocker = mocker
        self._capfd = capfd
        self._setup_files(self.path, base, files)

    @staticmethod
    def _setup_files(dest: Path, base: Path | None, content: dict[str, Any]) -> None:
        if base is not None:
            shutil.copytree(str(base), str(dest), ignore=shutil.ignore_patterns(".tox"))
        dest.mkdir(exist_ok=True)
        for key, value in content.items():
            if not isinstance(key, str):
                msg = f"{key!r} at {dest}"
                raise TypeError(msg)  # pragma: no cover
            at_path = dest / key
            if callable(value):
                value = textwrap.dedent("\n".join(inspect.getsourcelines(value)[0][1:]))  # noqa: PLW2901
                value = f"from __future__ import annotations\n{value}"  # noqa: PLW2901
            if isinstance(value, dict):
                at_path.mkdir(exist_ok=True)
                ToxProject._setup_files(at_path, None, value)
            elif isinstance(value, str):
                at_path.write_text(textwrap.dedent(value), encoding="utf-8")
            elif value is None:
                at_path.mkdir()
            else:
                msg = f"could not handle {at_path / key} with content {value!r}"  # pragma: no cover
                raise TypeError(msg)  # pragma: no cover

    def patch_execute(self, handle: Callable[[ExecuteRequest], int | None] | None = None) -> MagicMock:  # noqa: C901
        class MockExecute(Execute):
            def __init__(self, colored: bool, exit_code: int) -> None:  # noqa: FBT001
                self.exit_code = exit_code
                super().__init__(colored)

            def build_instance(
                self,
                request: ExecuteRequest,
                options: ExecuteOptions,
                out: SyncWrite,
                err: SyncWrite,
            ) -> ExecuteInstance:
                return MockExecuteInstance(request, options, out, err, self.exit_code)

        class MockExecuteStatus(ExecuteStatus):
            def __init__(self, options: ExecuteOptions, out: SyncWrite, err: SyncWrite, exit_code: int) -> None:
                super().__init__(options, out, err)
                self._exit_code = exit_code

            @property
            def exit_code(self) -> int | None:
                return self._exit_code

            def wait(self, timeout: float | None = None) -> int | None:  # noqa: ARG002
                return self._exit_code

            def write_stdin(self, content: str) -> None:  # noqa: ARG002, PLR6301
                return None  # pragma: no cover

            def interrupt(self) -> None:  # noqa: PLR6301
                return None  # pragma: no cover

        class MockExecuteInstance(ExecuteInstance):
            def __init__(
                self,
                request: ExecuteRequest,
                options: ExecuteOptions,
                out: SyncWrite,
                err: SyncWrite,
                exit_code: int,
            ) -> None:
                super().__init__(request, options, out, err)
                self.exit_code = exit_code

            def __enter__(self) -> ExecuteStatus:
                return MockExecuteStatus(self.options, self._out, self._err, self.exit_code)

            def __exit__(
                self,
                exc_type: type[BaseException] | None,
                exc_val: BaseException | None,
                exc_tb: TracebackType | None,
            ) -> None:
                pass

            @property
            def cmd(self) -> Sequence[str]:
                return self.request.cmd

        @contextmanager
        def _execute_call(
            self: ToxEnv,
            executor: Execute,
            out_err: OutErr,
            request: ExecuteRequest,
            show: bool,  # noqa: FBT001
        ) -> Iterator[ExecuteStatus]:
            exit_code = 0 if handle is None else handle(request)
            if exit_code is not None:
                executor = MockExecute(colored=executor._colored, exit_code=exit_code)  # noqa: SLF001
            with original_execute_call(self, executor, out_err, request, show) as status:
                yield status

        original_execute_call = ToxEnv._execute_call  # noqa: SLF001
        return self.mocker.patch.object(ToxEnv, "_execute_call", side_effect=_execute_call, autospec=True)

    @property
    def structure(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for dir_name, _, files in os.walk(str(self.path)):
            dir_path = Path(dir_name)
            into = result
            relative = dir_path.relative_to(str(self.path))
            for elem in relative.parts:
                into = into.setdefault(elem, {})
            for file_name in files:
                into[file_name] = (dir_path / file_name).read_text()
        return result

    @contextmanager
    def chdir(self, to: Path | None = None) -> Iterator[None]:
        cur_dir = Path.cwd()
        os.chdir(to or self.path)
        try:
            yield
        finally:
            os.chdir(cur_dir)

    def run(self, *args: str, from_cwd: Path | None = None, raise_on_config_fail: bool = True) -> ToxRunOutcome:
        if raise_on_config_fail and args and args[0] in {"c", "config"}:
            self.monkeypatch.setenv("_TOX_SHOW_CONFIG_RAISE", "1")

        with self.chdir(from_cwd):
            state = None
            self._capfd.readouterr()  # start with a clean state - drain
            code = None
            state = None

            def our_setup_state(value: Sequence[str]) -> State:
                nonlocal state
                state = previous_setup_state(value)
                return state

            with self.monkeypatch.context() as m:
                m.setattr(tox_env_api, "_CWD", self.path)
                m.setattr(tox.run, "setup_state", our_setup_state)
                m.setattr(sys, "argv", [sys.executable, "-m", "tox", *list(args)])
                m.setenv("VIRTUALENV_SYMLINK_APP_DATA", "1")
                m.setenv("VIRTUALENV_SYMLINKS", "1")
                m.setenv("VIRTUALENV_PIP", "embed")
                m.setenv("VIRTUALENV_SETUPTOOLS", "embed")
                try:
                    tox_run(args)
                except SystemExit as exception:
                    code = exception.code
                if code is None:  # pragma: no branch
                    msg = "exit code not set"
                    raise RuntimeError(msg)
            out, err = self._capfd.readouterr()
            return ToxRunOutcome(args, self.path, cast("int", code), out, err, state)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(path={self.path}) at {id(self)}"


@pytest.fixture(autouse=True, scope="session")
def enable_pep517_backend_coverage() -> Iterator[None]:
    try:
        import coverage  # noqa: F401, PLC0415
    except ImportError:  # pragma: no cover
        yield  # pragma: no cover
        return  # pragma: no cover
    # the COV_ env variables needs to be passed on for the PEP-517 backend
    from tox.tox_env.python.virtual_env.package.pyproject import Pep517VirtualEnvPackager  # noqa: PLC0415

    def default_pass_env(self: Pep517VirtualEnvPackager) -> list[str]:
        result = previous(self)
        result.append("COV_*")
        return result

    # monkey-patch to inject COV_* env vars, session-scoped so mocker unavailable
    previous = Pep517VirtualEnvPackager._default_pass_env  # noqa: SLF001
    try:
        Pep517VirtualEnvPackager._default_pass_env = default_pass_env  # noqa: SLF001  # ty: ignore[invalid-assignment]
        yield
    finally:
        Pep517VirtualEnvPackager._default_pass_env = previous  # noqa: SLF001


class ToxRunOutcome:
    def __init__(  # noqa: PLR0913
        self,
        cmd: Sequence[str],
        cwd: Path,
        code: int,
        out: str,
        err: str,
        state: State | None,
    ) -> None:
        extended_cmd = [sys.executable, "-m", "tox"]
        extended_cmd.extend(cmd)
        self.cmd: list[str] = extended_cmd
        self.cwd: Path = cwd
        self.code: int = code
        self.out: str = out
        self.err: str = err
        self._state: State | None = state

    @property
    def state(self) -> State:
        if self._state is None:
            msg = "no state"
            raise RuntimeError(msg)
        return self._state

    def env_conf(self, name: str) -> EnvConfigSet:
        return self.state.conf.get_env(name)

    @property
    def success(self) -> bool:
        return self.code == Outcome.OK

    def assert_success(self) -> None:
        assert self.success, repr(self)  # noqa: S101

    def assert_failed(self, code: int | None = None) -> None:
        status_match = self.code != 0 if code is None else self.code == code
        assert status_match, f"should be {code}, got {self}"  # noqa: S101

    def __repr__(self) -> str:
        return "\n".join(
            "{}{}{}".format(k, "\n" if "\n" in v else ": ", v)
            for k, v in (
                ("code", str(self.code)),
                ("cmd", self.shell_cmd),
                ("cwd", str(self.cwd)),
                ("standard output", self.out),
                ("standard error", self.err),
            )
            if v
        )

    @property
    def shell_cmd(self) -> str:
        return shell_cmd(self.cmd)

    def assert_out_err(self, out: str, err: str, *, dedent: bool = True, regex: bool = False) -> None:
        if dedent:
            out = textwrap.dedent(out).lstrip()
        if regex:
            self.matches(out, self.out, re.MULTILINE | re.DOTALL)
        else:
            assert self.out == out  # noqa: S101
        if dedent:
            err = textwrap.dedent(err).lstrip()
        if regex:
            self.matches(err, self.err, re.MULTILINE | re.DOTALL)
        else:
            assert self.err == err  # noqa: S101

    @staticmethod
    def matches(pattern: str, text: str, flags: int = 0) -> None:
        try:
            from re_assert import Matches  # noqa: PLC0415
        except ImportError:  # pragma: no cover # hard to test
            match = re.match(pattern, text, flags)
            if match is None:
                warnings.warn("install the re-assert PyPI package for bette error message", UserWarning, stacklevel=1)
            assert match  # noqa: S101
        else:
            assert Matches(pattern, flags=flags) == text  # noqa: S101


class ToxProjectCreator(Protocol):
    def __call__(
        self,
        files: dict[str, Any],
        base: Path | None = None,
        prj_path: Path | None = None,
    ) -> ToxProject: ...


@pytest.fixture(name="tox_project")
def init_fixture(
    tmp_path: Path,
    capfd: CaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    mocker: MockerFixture,
) -> ToxProjectCreator:
    def _init(files: dict[str, Any], base: Path | None = None, prj_path: Path | None = None) -> ToxProject:
        """Create tox projects."""
        return ToxProject(files, base, prj_path or tmp_path / "p", capfd, monkeypatch, mocker)

    return _init


@pytest.fixture
def empty_project(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> ToxProject:
    project = tox_project({"tox.ini": ""})
    monkeypatch.chdir(project.path)
    return project


_RUN_INTEGRATION_TEST_FLAG = "--run-integration"


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(_RUN_INTEGRATION_TEST_FLAG, action="store_true", help="run the integration tests")


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line("markers", "integration")
    config.addinivalue_line("markers", "plugin_test")


@pytest.hookimpl(trylast=True)
def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Function]) -> None:
    # do not require flags if called directly
    if len(items) == 1:  # pragma: no cover # hard to test
        return

    skip_int = pytest.mark.skip(reason=f"integration tests not run (no {_RUN_INTEGRATION_TEST_FLAG} flag)")

    def is_integration(test_item: pytest.Function) -> bool:
        return test_item.get_closest_marker("integration") is not None

    integration_enabled = config.getoption(_RUN_INTEGRATION_TEST_FLAG)
    if not integration_enabled:  # pragma: no cover # hard to test
        for item in items:
            if is_integration(item):
                item.add_marker(skip_int)
    # run integration tests (is_integration is True) after unit tests (False)
    items.sort(key=is_integration)


def enable_pypi_server(monkeypatch: pytest.MonkeyPatch, url: str | None) -> None:
    if url is None:  # pragma: no cover # only one of the branches can be hit depending on env
        monkeypatch.delenv("PIP_INDEX_URL", raising=False)
    else:  # pragma: no cover
        monkeypatch.setenv("PIP_INDEX_URL", url)
    monkeypatch.setenv("PIP_RETRIES", str(5))
    monkeypatch.setenv("PIP_TIMEOUT", str(2))


@pytest.fixture(scope="session")
def pypi_server(tmp_path_factory: pytest.TempPathFactory) -> Iterator[IndexServer]:
    # takes around 2.5s
    path = tmp_path_factory.mktemp("pypi")
    with IndexServer(path) as server:
        server.create_index("empty", "volatile=False")
        yield server


@pytest.fixture(scope="session")
def _invalid_index_fake_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as socket_handler:
        socket_handler.bind(("", 0))
        return cast("int", socket_handler.getsockname()[1])


@pytest.fixture(autouse=True)
def disable_pip_pypi_access(_invalid_index_fake_port: int, monkeypatch: pytest.MonkeyPatch) -> tuple[str, str | None]:
    """Set a fake pip index url, tests that want to use a pypi server should create and overwrite this."""
    previous_url = os.environ.get("PIP_INDEX_URL")
    new_url = f"http://localhost:{_invalid_index_fake_port}/bad-pypi-server"
    monkeypatch.setenv("PIP_INDEX_URL", new_url)
    monkeypatch.setenv("PIP_RETRIES", str(0))
    monkeypatch.setenv("PIP_TIMEOUT", str(0.001))
    return new_url, previous_url


@pytest.fixture(name="enable_pip_pypi_access")
def enable_pip_pypi_access_fixture(
    disable_pip_pypi_access: tuple[str, str | None],
    monkeypatch: pytest.MonkeyPatch,
) -> str | None:
    """Set a fake pip index url, tests that want to use a pypi server should create and overwrite this."""
    _, previous_url = disable_pip_pypi_access
    enable_pypi_server(monkeypatch, previous_url)
    return previous_url


def register_inline_plugin(mocker: MockerFixture, *args: Callable[..., Any]) -> None:
    frame_info = inspect.stack()[1]
    caller_module = inspect.getmodule(frame_info[0])
    assert caller_module is not None  # noqa: S101
    plugin = ModuleType(f"{caller_module.__name__}|{frame_info[3]}")
    plugin.__file__ = caller_module.__file__
    plugin.__dict__.update({f.__name__: f for f in args})  # ty: ignore[unresolved-attribute] # https://github.com/astral-sh/ty/issues/1495
    mocker.patch("tox.plugin.manager.load_inline", return_value=plugin)


LogCaptureFixture = pytest.LogCaptureFixture
TempPathFactory = pytest.TempPathFactory
MonkeyPatch = pytest.MonkeyPatch

__all__ = (
    "CaptureFixture",
    "LogCaptureFixture",
    "MonkeyPatch",
    "SubRequest",
    "TempPathFactory",
    "ToxProject",
    "ToxProjectCreator",
    "ToxRunOutcome",
    "check_os_environ",
    "register_inline_plugin",
)



from __future__ import annotations

from typing import TYPE_CHECKING

import argcomplete

from tox.config.cli.parse import get_options

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_argcomplete_autocomplete_called(mocker: MockerFixture) -> None:
    mock_autocomplete = mocker.patch.object(argcomplete, "autocomplete")
    get_options("r")
    mock_autocomplete.assert_called_once()


def test_argcomplete_missing_does_not_break(mocker: MockerFixture) -> None:
    mocker.patch.dict("sys.modules", {"argcomplete": None})
    result = get_options("r")
    assert result.parsed is not None



from __future__ import annotations

import ast
import glob
import os
import re
import sys
import sysconfig
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from tox.config.loader.ini.factor import find_factor_groups
from tox.config.loader.replacer import (
    MatchError,
    MatchRecursionError,
    ReplaceReference,
    load_posargs,
    replace,
    replace_env,
)
from tox.config.loader.stringify import stringify
from tox.config.types import Command

from ._validate import validate

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tox.config.loader.api import ConfigLoadArgs
    from tox.config.loader.toml import TomlLoader
    from tox.config.main import Config
    from tox.config.sets import ConfigSet
    from tox.config.source.toml_pyproject import TomlSection

    from ._api import TomlTypes


class Unroll:
    def __init__(self, conf: Config | None, loader: TomlLoader, args: ConfigLoadArgs) -> None:
        self.conf = conf
        self.loader = loader
        self.args = args
        self.factors = self._extract_factors(args.env_name)

    @staticmethod
    def _extract_factors(env_name: str | None) -> set[str]:
        """Extract factors from environment name and add platform."""
        if env_name is None:
            factors = set()
        else:
            factors = set(chain.from_iterable([(i for i, _ in a) for a in find_factor_groups(env_name)]))
        factors.add(sys.platform)
        parts = sysconfig.get_platform().rsplit("-", 1)
        if len(parts) > 1:
            factors.add(parts[-1])
        return factors

    def __call__(  # noqa: C901, PLR0912
        self, value: TomlTypes, depth: int = 0, *, skip_str: bool = False
    ) -> TomlTypes:
        """Replace all active tokens within value according to the config."""
        depth += 1
        MatchRecursionError.check(depth, value)
        if isinstance(value, str):
            if not skip_str and self.conf is not None:  # core config does not support string substitution
                reference = TomlReplaceLoader(self.conf, self.loader)
                value = replace(self.conf, reference, value, self.args)
        elif isinstance(value, (int, float, bool)):
            pass  # no reference or substitution possible
        elif isinstance(value, list):
            # need to inspect every entry of the list to check for reference.
            res_list: list[TomlTypes] = []
            for val in value:  # apply replacement for every entry
                got = self(val, depth, skip_str=skip_str)
                if isinstance(val, dict) and val.get("replace") and val.get("extend"):
                    res_list.extend(cast("list[Any]", got))
                else:
                    res_list.append(got)
            value = res_list
        elif isinstance(value, dict):
            # need to inspect every entry of the list to check for reference.
            if replace_type := value.get("replace"):
                marker = value.get("marker")
                if replace_type == "posargs" and self.conf is not None:
                    got_posargs = load_posargs(self.conf, self.args)
                    posargs_result: TomlTypes = (
                        [self(v, depth, skip_str=skip_str) for v in cast("list[str]", value.get("default", []))]
                        if got_posargs is None
                        else list(got_posargs)
                    )
                    return {"value": posargs_result, "marker": marker} if marker else posargs_result
                if replace_type == "env":
                    env_result: TomlTypes = replace_env(
                        self.conf,
                        [
                            validate(value["name"], str),
                            validate(self(value.get("default", ""), depth, skip_str=skip_str), str),
                        ],
                        self.args,
                    )
                    return {"value": env_result, "marker": marker} if marker else env_result
                if replace_type == "glob":
                    glob_result = _replace_glob_toml(self.conf, value)
                    return {"value": glob_result, "marker": marker} if marker else glob_result
                if replace_type == "if":
                    if_result = _replace_if_toml(value, self, depth, self.factors, skip_str=skip_str)
                    return {"value": if_result, "marker": marker} if marker else if_result
                if replace_type == "ref":  # pragma: no branch
                    ref_result = self._replace_ref(value, depth, skip_str=skip_str)
                    return {"value": ref_result, "marker": marker} if marker else ref_result

            res_dict: dict[str, TomlTypes] = {}
            for key, val in value.items():  # apply replacement for every entry
                res_dict[key] = self(val, depth, skip_str=skip_str)
            value = res_dict
        return value

    def _replace_ref(self, value: dict[str, TomlTypes], depth: int, *, skip_str: bool = False) -> TomlTypes:
        if self.conf is not None and (env := value.get("env")) and (key := value.get("key")):
            result = self.conf.get_env(cast("str", env))[cast("str", key)]
            if isinstance(result, Command):
                return cast("TomlTypes", result.args)
            return cast("TomlTypes", result)
        if of := value.get("of"):
            validated_of = validate(of, list[str])
            loaded = self.loader.load_raw_from_root(self.loader.section.SEP.join(validated_of))
            return self(loaded, depth, skip_str=skip_str)
        return value


def _replace_glob_toml(conf: Config | None, value: dict[str, Any]) -> list[str] | str:
    pattern = validate(value.get("pattern"), str)
    if not pattern:
        msg = "No pattern was supplied in glob replacement"
        raise MatchError(msg)
    if conf is not None and not Path(pattern).is_absolute():
        pattern = str(conf.core["tox_root"] / pattern)
    extending = value.get("extend", False)
    if matches := sorted(glob.glob(pattern, recursive=True)):  # noqa: PTH207
        return matches if extending else " ".join(matches)
    default = value.get("default")
    if default is None:
        return [] if extending else ""
    return validate(default, list) if extending else validate(default, str)


def _replace_if_toml(
    value: dict[str, Any], unroll: Unroll, depth: int, factors: set[str], *, skip_str: bool = False
) -> TomlTypes:
    condition = value.get("condition")
    if not condition or not isinstance(condition, str):
        msg = "No condition was supplied in if replacement"
        raise MatchError(msg)
    if "then" not in value:
        msg = "No 'then' value was supplied in if replacement"
        raise MatchError(msg)
    return unroll(
        value["then"] if _evaluate_condition(condition, factors) else value.get("else", ""), depth, skip_str=skip_str
    )


def _evaluate_condition(expr: str, factors: set[str]) -> bool:  # noqa: C901
    """Evaluate a condition expression supporting env.VAR, factor.NAME lookups, comparisons, and boolean logic."""
    try:
        tree = ast.parse(expr, mode="eval")
    except SyntaxError:
        msg = f"Invalid condition expression: {expr}"
        raise MatchError(msg) from None

    def _eval(node: ast.expr) -> str | bool:  # noqa: PLR0911, C901
        if isinstance(node, ast.BoolOp):
            if isinstance(node.op, ast.And):
                return all(bool(_eval(v)) for v in node.values)
            return any(bool(_eval(v)) for v in node.values)
        if isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
            return not bool(_eval(node.operand))
        if isinstance(node, ast.Compare) and len(node.ops) == 1 and len(node.comparators) == 1:
            left = _eval(node.left)
            right = _eval(node.comparators[0])
            if isinstance(node.ops[0], ast.Eq):
                return left == right
            if isinstance(node.ops[0], ast.NotEq):
                return left != right
            msg = f"Unsupported comparison operator in condition: {expr}"
            raise MatchError(msg)
        if isinstance(node, ast.Attribute) and isinstance(node.value, ast.Name):
            if node.value.id == "env":
                return os.environ.get(node.attr, "")
            if node.value.id == "factor":
                return node.attr in factors
            msg = f"Unsupported namespace in condition: {node.value.id} (expected 'env' or 'factor')"
            raise MatchError(msg)
        if isinstance(node, ast.Constant) and isinstance(node.value, str):
            return node.value
        msg = f"Unsupported expression in condition: {ast.dump(node)}"
        raise MatchError(msg)

    return bool(_eval(tree.body))


_REFERENCE_PATTERN = re.compile(
    r"""
    (\[(?P<section>.*)])? # default value
    (?P<key>[-a-zA-Z0-9_]+) # key
    (:(?P<default>.*))? # default value
    $
""",
    re.VERBOSE,
)


class TomlReplaceLoader(ReplaceReference):
    def __init__(self, conf: Config, loader: TomlLoader) -> None:
        self.conf = conf
        self.loader = loader

    def __call__(self, value: str, conf_args: ConfigLoadArgs) -> str | None:
        if match := _REFERENCE_PATTERN.search(value):
            settings = match.groupdict()
            exception: Exception | None = None
            try:
                for src in self._config_value_sources(settings["section"], conf_args.env_name):
                    try:
                        value = src.load(settings["key"], conf_args.chain)
                    except KeyError as exc:  # if fails, keep trying maybe another source can satisfy # noqa: PERF203
                        exception = exc
                    else:
                        return stringify(value)[0]
            except Exception as exc:  # noqa: BLE001
                exception = exc
            if exception is not None:
                if isinstance(exception, KeyError):  # if the lookup failed replace - else keep
                    default = settings["default"]
                    if default is not None:
                        return default
                    return None  # keep original text, consistent with ini loader behavior
                raise exception
        return value

    def _config_value_sources(self, sec: str | None, current_env: str | None) -> Iterator[ConfigSet | RawLoader]:
        if sec is None:
            if current_env is not None:  # pragma: no branch
                yield self.conf.get_env(current_env)
            yield self.conf.core
            return

        section = cast("TomlSection", self.loader.section)
        core_prefix = section.core_prefix()
        env_prefix = section.env_prefix()
        run_env_base = section.run_env_base()
        pkg_env_base = section.package_env_base()
        if sec.startswith(env_prefix) and sec not in {run_env_base, pkg_env_base}:
            env = sec[len(env_prefix) + len(section.SEP) :]
            yield self.conf.get_env(env)
        else:
            yield RawLoader(self.loader, sec)
        if sec == core_prefix:
            yield self.conf.core  # try via registered configs


class RawLoader:
    def __init__(self, loader: TomlLoader, section: str) -> None:
        self._loader = loader
        self._section = section

    def load(self, item: str, chain: list[str] | None = None) -> Any:  # noqa: ARG002
        return self._loader.load_raw_from_root(f"{self._section}{self._loader.section.SEP}{item}")


__all__ = [
    "TomlReplaceLoader",
    "Unroll",
]



"""Group together configuration values (such as base tox configuration, tox environment configs)."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar, cast

from tox.config.loader.api import ConfigLoadArgs, Loader
from tox.config.types import CircularChainError

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from types import UnionType

    from tox.config.loader.convert import Factory
    from tox.config.main import Config  # pragma: no cover


T = TypeVar("T")
V = TypeVar("V")


class ConfigDefinition(ABC, Generic[T]):  # noqa: PLW1641
    """Abstract base class for configuration definitions."""

    def __init__(self, keys: Iterable[str], desc: str) -> None:
        self.keys = keys
        self.desc = desc

    @abstractmethod
    def __call__(self, conf: Config, loaders: list[Loader[T]], args: ConfigLoadArgs) -> T:
        raise NotImplementedError

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ConfigDefinition):
            return False
        return (self.keys, self.desc) == (o.keys, o.desc)

    def __ne__(self, o: object) -> bool:
        return not (self == o)


class ConfigConstantDefinition(ConfigDefinition[T]):  # noqa: PLW1641
    """A configuration definition whose value is defined upfront (such as the tox environment name)."""

    def __init__(
        self,
        keys: Iterable[str],
        desc: str,
        value: Callable[[], T] | T,
    ) -> None:
        super().__init__(keys, desc)
        self.value = value

    def __call__(
        self,
        conf: Config,  # noqa: ARG002
        loaders: list[Loader[T]],  # noqa: ARG002
        args: ConfigLoadArgs,  # noqa: ARG002
    ) -> T:
        return self.value() if callable(self.value) else self.value  # ty: ignore[call-top-callable]

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ConfigConstantDefinition):
            return False
        return super().__eq__(o) and self.value == o.value

    def __repr__(self) -> str:
        values = ((k, v) for k, v in vars(self).items() if v is not None)
        return f"{type(self).__name__}({', '.join(f'{k}={v}' for k, v in values)})"


_PLACE_HOLDER = object()


class ConfigDynamicDefinition(ConfigDefinition[T]):  # noqa: PLW1641
    """A configuration definition that comes from a source (such as in memory, an ini file, a toml file, etc.)."""

    def __init__(  # noqa: PLR0913
        self,
        keys: Iterable[str],
        desc: str,
        of_type: type[T] | UnionType,
        default: Callable[[Config, str | None], T] | T,
        post_process: Callable[[T], T] | None = None,
        factory: Factory[T] | None = None,
    ) -> None:
        super().__init__(keys, desc)
        self.of_type = of_type
        self.default = default
        self.post_process = post_process
        self.factory = factory
        self._cache: object | T = _PLACE_HOLDER

    def __call__(
        self,
        conf: Config,
        loaders: list[Loader[T]],
        args: ConfigLoadArgs,
    ) -> T:
        if self._cache is _PLACE_HOLDER:
            primary_key, *alias_keys = self.keys
            for loader in loaders:
                chain_key = f"{loader.section.key}.{primary_key}"
                try:
                    if chain_key in args.chain:
                        values = args.chain[args.chain.index(chain_key) :]
                        msg = f"circular chain detected {', '.join(values)}"
                        raise CircularChainError(msg)
                finally:
                    args.chain.append(chain_key)
                try:
                    value = loader.load(primary_key, self.of_type, self.factory, conf, args, all_keys=alias_keys)
                except KeyError:
                    continue
                else:
                    break
                finally:
                    del args.chain[-1]
            else:
                value = self.default(conf, args.env_name) if callable(self.default) else self.default  # ty: ignore[call-top-callable]
            if self.post_process is not None:
                value = self.post_process(value)
            self._cache = value
        return cast("T", self._cache)

    def __repr__(self) -> str:
        values = ((k, v) for k, v in vars(self).items() if k not in {"post_process", "_cache"} and v is not None)
        return f"{type(self).__name__}({', '.join(f'{k}={v}' for k, v in values)})"

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ConfigDynamicDefinition):
            return False
        return super().__eq__(o) and (self.of_type, self.default, self.post_process) == (
            o.of_type,
            o.default,
            o.post_process,
        )


__all__ = [
    "ConfigConstantDefinition",
    "ConfigDefinition",
    "ConfigDynamicDefinition",
    "ConfigLoadArgs",
]



"""Execute a command in a tox environment."""

from __future__ import annotations

from typing import TYPE_CHECKING

from tox.config.loader.memory import MemoryLoader
from tox.config.types import Command
from tox.plugin import impl
from tox.report import HandledError
from tox.session.cmd.run.common import env_run_create_flags
from tox.session.cmd.run.sequential import run_sequential
from tox.session.env_select import CliEnv, register_env_select_flags

if TYPE_CHECKING:
    from pathlib import Path

    from tox.config.cli.parser import ToxParser
    from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("exec", ["e"], "execute an arbitrary command within a tox environment", exec_)
    our.epilog = "For example: tox exec -e py39 -- python --version"
    register_env_select_flags(our, default=CliEnv("py"), multiple=False)
    env_run_create_flags(our, mode="exec")


def exec_(state: State) -> int:
    state.conf.options.skip_pkg_install = True  # avoid package install
    state.conf.options.no_capture = True  # always run interactively without output capture
    envs = list(state.envs.iter())
    if len(envs) != 1:
        msg = f"exactly one target environment allowed in exec mode but found {', '.join(envs)}"
        raise HandledError(msg)
    loader = MemoryLoader(  # these configuration values are loaded from in-memory always (no file conf)
        commands_pre=[],
        commands=[],
        commands_post=[],
    )
    conf = state.envs[envs[0]].conf
    conf.loaders.insert(0, loader)
    to_path: Path | None = conf["change_dir"] if conf["args_are_paths"] else None
    pos_args = state.conf.pos_args(to_path)
    if not pos_args:
        msg = "You must specify a command as positional arguments, use -- <command>"
        raise HandledError(msg)
    loader.raw["commands"] = [Command(list(pos_args))]
    return run_sequential(state)



from __future__ import annotations

import sys
from collections import defaultdict
from typing import TYPE_CHECKING, TypedDict

from packaging.requirements import InvalidRequirement, Requirement
from packaging.utils import canonicalize_name

from tox.tox_env.errors import Fail

if TYPE_CHECKING:
    from pathlib import Path


if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    import tomllib
else:  # pragma: <3.11 cover
    import tomli as tomllib

_IncludeGroup = TypedDict("_IncludeGroup", {"include-group": str})


def _add_extra_to_deps(
    dependency_groups: dict[str, list[str]],
    dependencies: set[Requirement],
    extra: str,
    seen_extras: set[str],
) -> None:
    """Add dependencies for a given extra to the dependencies set."""
    normed_extra = canonicalize_name(extra)
    if normed_extra in seen_extras:
        return
    seen_extras.add(normed_extra)
    if normed_extra not in dependency_groups:
        msg = f"extra {extra!r} not found in dependency groups"
        raise Fail(msg)
    for dep_str in dependency_groups[normed_extra]:
        try:
            dependencies.add(Requirement(dep_str))
        except InvalidRequirement as exc:  # noqa: PERF203
            msg = f"{dep_str!r} is not valid requirement due to {exc}"
            raise Fail(msg) from exc


def unwrap_nested_extras(
    dependency_groups: dict[str, list[str]],
    project_name: str | None,
    dependencies: set[Requirement],
    seen_extras: set[str],
) -> set[Requirement]:
    """Unwrap nested dependency groups into a flat set of dependencies."""
    if not project_name:
        return dependencies

    extras_to_unwrap: set[Requirement] = set()
    for dependency in dependencies:
        if dependency.name == project_name:
            extras_to_unwrap.add(dependency)
    if not extras_to_unwrap:
        return dependencies

    for dependency in extras_to_unwrap:
        dependencies.remove(dependency)
        for extra in dependency.extras:
            _add_extra_to_deps(dependency_groups, dependencies, extra, seen_extras)
    return unwrap_nested_extras(dependency_groups, project_name, dependencies, seen_extras)


def resolve(root: Path, groups: set[str]) -> set[Requirement]:
    pyproject_file = root / "pyproject.toml"
    if not pyproject_file.exists():  # check if it's static PEP-621 metadata
        return set()
    with pyproject_file.open("rb") as file_handler:
        pyproject = tomllib.load(file_handler)
    dependency_groups_raw = pyproject["dependency-groups"]
    if not isinstance(dependency_groups_raw, dict):
        msg = f"dependency-groups is {type(dependency_groups_raw).__name__} instead of table"
        raise Fail(msg)
    original_names_lookup, dependency_groups = _normalize_group_names(dependency_groups_raw)
    result: set[Requirement] = set()
    for group in groups:
        result = result.union(_resolve_dependency_group(dependency_groups, group, original_names_lookup))

    project_name = pyproject.get("project", {}).get("name")
    optional_dependencies = pyproject.get("project", {}).get("optional-dependencies", {})

    return unwrap_nested_extras(optional_dependencies, project_name, result, set())


def _normalize_group_names(
    dependency_groups: dict[str, list[str] | _IncludeGroup],
) -> tuple[dict[str, str], dict[str, list[str] | _IncludeGroup]]:
    original_names = defaultdict(list)
    normalized_groups = {}

    for group_name, value in dependency_groups.items():
        normed_group_name: str = canonicalize_name(group_name)
        original_names[normed_group_name].append(group_name)
        normalized_groups[normed_group_name] = value

    errors = []
    for normed_name, names in original_names.items():
        if len(names) > 1:
            errors.append(f"{normed_name} ({', '.join(names)})")
    if errors:
        msg = f"Duplicate dependency group names: {', '.join(errors)}"
        raise ValueError(msg)

    original_names_lookup = {
        normed_name: original_names[0]
        for normed_name, original_names in original_names.items()
        if len(original_names) == 1
    }

    return original_names_lookup, normalized_groups


def _resolve_dependency_group(
    dependency_groups: dict[str, list[str] | _IncludeGroup],
    group: str,
    original_names_lookup: dict[str, str],
    past_groups: tuple[str, ...] = (),
) -> set[Requirement]:
    if group in past_groups:
        original_group = original_names_lookup.get(group, group)
        original_past_groups = tuple(original_names_lookup.get(g, g) for g in past_groups)
        msg = f"Cyclic dependency group include: {original_group!r} -> {original_past_groups!r}"
        raise Fail(msg)
    if group not in dependency_groups:
        original_group = original_names_lookup.get(group, group)
        msg = f"dependency group {original_group!r} not found"
        raise Fail(msg)
    raw_group = dependency_groups[group]
    if not isinstance(raw_group, list):
        original_group = original_names_lookup.get(group, group)
        msg = f"dependency group {original_group!r} is not a list"
        raise Fail(msg)

    result = set()
    for item in raw_group:
        if isinstance(item, str):
            # packaging.requirements.Requirement parsing ensures that this is a valid
            # PEP 508 Dependency Specifier
            # raises InvalidRequirement on failure
            try:
                result.add(Requirement(item))
            except InvalidRequirement as exc:
                msg = f"{item!r} is not valid requirement due to {exc}"
                raise Fail(msg) from exc
        elif isinstance(item, dict) and tuple(item.keys()) == ("include-group",):
            include_group = canonicalize_name(next(iter(item.values())))
            result = result.union(
                _resolve_dependency_group(
                    dependency_groups, include_group, original_names_lookup, (*past_groups, group)
                )
            )
        else:
            msg = f"invalid dependency group item: {item!r}"
            raise Fail(msg)
    return result


__all__ = [
    "resolve",
]



from __future__ import annotations

import sys
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from tox.config.loader.ini import IniLoader
from tox.config.loader.ini.factor import LATEST_PYTHON_MINOR_MAX, LATEST_PYTHON_MINOR_MIN, filter_for_env, find_envs
from tox.config.source.ini_section import IniSection

if TYPE_CHECKING:
    from collections.abc import Callable
    from configparser import ConfigParser

    from tests.conftest import ToxIniCreator
    from tox.config.main import Config


def test_factor_env_discover_empty() -> None:
    result = list(find_envs("\n\n"))
    assert result == []


@pytest.fixture(scope="session")
def complex_example() -> str:
    return dedent(
        """
    default
    lines
    py: py only
    !py: not py
    {py,!pi}-{a,b}{,-dev},c: complex
    py, d: space
    extra: extra
    more-default
    no:space
    trailingcolon:
    tab:\ttab
    """,
    )


def test_factor_env_discover(complex_example: str) -> None:
    result = list(find_envs(complex_example))
    assert result == [
        "py",
        "py-a",
        "py-a-dev",
        "py-b",
        "py-b-dev",
        "pi-a",
        "pi-a-dev",
        "pi-b",
        "pi-b-dev",
        "c",
        "d",
        "extra",
        "trailingcolon",
        "tab",
    ]


@pytest.mark.parametrize(
    "env",
    [
        "py",
        "py-a",
        "py-a-dev",
        "py-b",
        "py-b-dev",
        "pi-a",
        "pi-a-dev",
        "pi-b",
        "pi-b-dev",
        "c",
        "extra",
        "trailingcolon",
        "tab",
    ],
)
def test_factor_env_filter(env: str, complex_example: str) -> None:
    result = filter_for_env(complex_example, name=env)
    assert "default" in result
    assert "lines" in result
    assert "more-default" in result
    assert "no:space" in result
    if "py" in env:
        assert "py only" in result
        assert "not py" not in result
    else:
        assert "py only" not in result
        assert "not py" in result
    if env == "extra":
        assert "extra" in result
    else:
        assert "extra" not in result
    if env in {"py-a", "py-a-dev", "py-b", "py-b-dev", "c"}:
        assert "complex" in result
    else:
        assert "complex" not in result


def test_factor_env_list(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf("[tox]\nenv_list = {py27,py36}-django{ 15, 16 }{,-dev}, docs, flake")
    result = list(config)
    assert result == [
        "py27-django15",
        "py27-django15-dev",
        "py27-django16",
        "py27-django16-dev",
        "py36-django15",
        "py36-django15-dev",
        "py36-django16",
        "py36-django16-dev",
        "docs",
        "flake",
    ]


def test_simple_env_list(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf("[tox]\nenv_list = docs, flake8")
    assert list(config) == ["docs", "flake8"]


def test_factor_config(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [tox]
        env_list = {py36,py37}-{django15,django16}
        [testenv]
        deps-x =
            pytest
            django15: Django>=1.5,<1.6
            django16: Django>=1.6,<1.7
            py36: unittest2
            !py37,!django16: negation-or
            !py37-!django16: negation-and
        """,
    )
    assert list(config) == ["py36-django15", "py36-django16", "py37-django15", "py37-django16"]
    for env in config.core["env_list"]:
        env_config = config.get_env(env)
        env_config.add_config(keys="deps-x", of_type=list[str], default=[], desc="deps")
        deps = env_config["deps-x"]
        assert "pytest" in deps
        if "py36" in env:
            assert "unittest2" in deps
            assert "negation-or" in deps
        if "django15" in env:
            assert "Django>=1.5,<1.6" in deps
            assert "negation-or" in deps
        if "django16" in env:
            assert "Django>=1.6,<1.7" in deps
        if env_config.name == "py36-django15":
            assert "negation-and" in deps


def test_factor_config_do_not_replace_unescaped_comma(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf("[tox]\nenv_list = py37-{base,i18n},b")
    assert list(config) == ["py37-base", "py37-i18n", "b"]


def test_factor_config_no_env_list_creates_env(tox_ini_conf: ToxIniCreator) -> None:
    """If we have a factor that is not specified within the core env-list then that's also an environment"""
    config = tox_ini_conf(
        """
        [tox]
        env_list = py37-{django15,django16}
        [testenv]
        deps =
            pytest
            django15: Django>=1.5,<1.6
            django16: Django>=1.6,<1.7
            py36: unittest2
        """,
    )

    assert list(config) == ["py37-django15", "py37-django16", "py36"]


@pytest.mark.parametrize(
    ("env_list", "expected_envs"),
    [
        pytest.param("py3{10-13}", ["py310", "py311", "py312", "py313"], id="Expand positive range"),
        pytest.param("py3{10-11},a", ["py310", "py311", "a"], id="Expand range and add additional env"),
        pytest.param("py3{10-11},a{1-2}", ["py310", "py311", "a1", "a2"], id="Expand multiple env with ranges"),
        pytest.param(
            "py3{10-12,14}",
            ["py310", "py311", "py312", "py314"],
            id="Expand ranges, and allow extra parameter in generator",
        ),
        pytest.param(
            "py3{8-10,12,14-16}",
            ["py38", "py39", "py310", "py312", "py314", "py315", "py316"],
            id="Expand multiple ranges for one generator",
        ),
        pytest.param(
            "py3{10-11}-django1.{3-5}",
            [
                "py310-django1.3",
                "py310-django1.4",
                "py310-django1.5",
                "py311-django1.3",
                "py311-django1.4",
                "py311-django1.5",
            ],
            id="Expand ranges and factor multiple environment parts",
        ),
        pytest.param(
            "py3{10-11, 13}-django1.{3-4, 6}",
            [
                "py310-django1.3",
                "py310-django1.4",
                "py310-django1.6",
                "py311-django1.3",
                "py311-django1.4",
                "py311-django1.6",
                "py313-django1.3",
                "py313-django1.4",
                "py313-django1.6",
            ],
            id="Expand ranges and parameters and factor multiple environment parts",
        ),
        pytest.param(
            "py3{10-11},a{1-2}-b{3-4}",
            ["py310", "py311", "a1-b3", "a1-b4", "a2-b3", "a2-b4"],
            id="Expand ranges and parameters & factor multiple environment parts for multiple generative environments",
        ),
        pytest.param("py3{13-11}", ["py313", "py312", "py311"], id="Expand negative ranges"),
        pytest.param("3.{10-13}", ["3.10", "3.11", "3.12", "3.13"], id="Expand new-style python envs"),
        pytest.param(
            "py3{9-}",
            [f"py3{v}" for v in range(9, LATEST_PYTHON_MINOR_MAX + 1)],
            id="Expand right-open range to LATEST_PYTHON_MINOR_MAX",
        ),
        pytest.param(
            "3.{10-}",
            [f"3.{v}" for v in range(10, LATEST_PYTHON_MINOR_MAX + 1)],
            id="Expand right-open range new-style envs",
        ),
        pytest.param(
            "py3{-13}",
            [f"py3{v}" for v in range(LATEST_PYTHON_MINOR_MIN, 14)],
            id="Expand left-open range from LATEST_PYTHON_MINOR_MIN",
        ),
        pytest.param(
            "foo{11-}",
            [f"foo{v}" for v in range(11, LATEST_PYTHON_MINOR_MAX + 1)],
            id="Expand right-open numerical range",
        ),
        pytest.param(
            "py3{-11}",
            [f"py3{v}" for v in range(LATEST_PYTHON_MINOR_MIN, 12)],
            id="Expand left-open numerical range",
        ),
        pytest.param("foo{a-}", ["fooa-"], id="Don't expand right-open range"),
        pytest.param("foo{-a}", ["foo-a"], id="Don't expand left-open range"),
        pytest.param("foo{a-11}", ["fooa-11"], id="Don't expand alpha-umerical range"),
        pytest.param("foo{13-a}", ["foo13-a"], id="Don't expand numerical-alpha range"),
        pytest.param("foo{a-b}", ["fooa-b"], id="Don't expand non-numerical range"),
    ],
)
def test_env_list_expands_ranges(env_list: str, expected_envs: list[str], tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        f"""
        [tox]
        env_list = {env_list}
        """
    )

    assert list(config) == expected_envs


@pytest.mark.parametrize(
    ("env", "result"),
    [
        ("py35", "python -m coverage html -d cov"),
        ("py36", "python -m coverage html -d cov\n--show-contexts"),
    ],
)
def test_ini_loader_raw_with_factors(
    mk_ini_conf: Callable[[str], ConfigParser],
    env: str,
    result: str,
    empty_config: Config,
) -> None:
    commands = "python -m coverage html -d cov \n    !py35: --show-contexts"
    loader = IniLoader(
        section=IniSection(None, "testenv"),
        parser=mk_ini_conf(f"[tox]\nenvlist=py35,py36\n[testenv]\ncommands={commands}"),
        overrides=[],
        core_section=IniSection(None, "tox"),
    )
    outcome = loader.load_raw(key="commands", conf=empty_config, env_name=env)
    assert outcome == result


@pytest.mark.parametrize(
    ("env", "result"),
    [
        ("foo", "python -c \"print('foo')\"\npython -c \"print('bar')\""),
        ("bar", "python -c \"print('bar')\""),
    ],
)
def test_ini_loader_factor_multiline_command(
    mk_ini_conf: Callable[[str], ConfigParser],
    env: str,
    result: str,
    empty_config: Config,
) -> None:
    commands = "foo: python -c \"\\\n        print('foo')\"\n    python -c \"print('bar')\""
    loader = IniLoader(
        section=IniSection(None, "testenv"),
        parser=mk_ini_conf(f"[tox]\nenvlist=foo,bar\n[testenv]\ncommands={commands}"),
        overrides=[],
        core_section=IniSection(None, "tox"),
    )
    outcome = loader.load_raw(key="commands", conf=empty_config, env_name=env)
    assert outcome == result


@pytest.mark.parametrize(
    ("env", "result"),
    [
        ("py-cov", "coverage run somefile.py"),
        ("py-no_cov", "python somefile.py"),
    ],
)
def test_ini_loader_factor_conditional_continuation(
    mk_ini_conf: Callable[[str], ConfigParser],
    env: str,
    result: str,
    empty_config: Config,
) -> None:
    commands = "cov: coverage run \\\n    !cov: python \\\n        somefile.py"
    loader = IniLoader(
        section=IniSection(None, "testenv"),
        parser=mk_ini_conf(f"[tox]\nenvlist=py-cov,py-no_cov\n[testenv]\ncommands={commands}"),
        overrides=[],
        core_section=IniSection(None, "tox"),
    )
    outcome = loader.load_raw(key="commands", conf=empty_config, env_name=env)
    assert outcome == result


@pytest.mark.parametrize(
    ("env", "result"),
    [
        ("py312", "pytest --remote-data --durations=10"),
        ("py312-coverage", "coverage run -m pytest --remote-data --durations=10"),
    ],
)
def test_ini_loader_factor_mixed_continuation(
    mk_ini_conf: Callable[[str], ConfigParser],
    env: str,
    result: str,
    empty_config: Config,
) -> None:
    ini = dedent("""\
        [tox]
        envlist = py312,py312-coverage,py312-devdeps,py312-compatibility,py312-mocks3
        [testenv]
        commands =
            coverage: coverage run -m \\
            pytest \\
            devdeps: -W some_warning
            compatibility: integration_tests/ \\
            mocks3: tests/ \\
            --remote-data \\
            --durations=10
        """)
    loader = IniLoader(
        section=IniSection(None, "testenv"),
        parser=mk_ini_conf(ini),
        overrides=[],
        core_section=IniSection(None, "tox"),
    )
    outcome = loader.load_raw(key="commands", conf=empty_config, env_name=env)
    assert outcome == result


def test_generative_ranges_in_deps(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [testenv]
        deps-x =
            py{310-314}: black
        """,
    )
    assert list(config) == ["py310", "py311", "py312", "py313", "py314"]


def test_generative_ranges_in_deps_with_mixed_approach(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [testenv]
        deps-x =
            py3{10-14}: black
        """,
    )
    assert list(config) == ["py310", "py311", "py312", "py313", "py314"]


def test_generative_ranges_in_setenv(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [testenv]
        setenv =
            foo{1,2}: FOO=bar
            foo{3-5}: FOO=baz
        """,
    )
    assert list(config) == ["foo1", "foo2", "foo3", "foo4", "foo5"]


def test_generative_section_name_with_ranges(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [testenv:py3{11-13}-{black,lint}]
        deps-x =
            black: black
            lint: flake8
        """,
    )
    assert list(config) == ["py311-black", "py311-lint", "py312-black", "py312-lint", "py313-black", "py313-lint"]


def test_generative_section_name(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [testenv:{py311,py310}-{black,lint}]
        deps-x =
            black: black
            lint: flake8
        """,
    )
    assert list(config) == ["py311-black", "py311-lint", "py310-black", "py310-lint"]

    env_config = config.get_env("py311-black")
    env_config.add_config(keys="deps-x", of_type=list[str], default=[], desc="deps")
    deps = env_config["deps-x"]
    assert deps == ["black"]

    env_config = config.get_env("py311-lint")
    env_config.add_config(keys="deps-x", of_type=list[str], default=[], desc="deps")
    deps = env_config["deps-x"]
    assert deps == ["flake8"]


def test_multiple_factor_match(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf("[testenv]\nconf = a{,-b}: x")
    env_config = config.get_env("a-b")
    env_config.add_config(keys="conf", of_type=str, default="", desc="conf")
    deps = env_config["conf"]
    assert deps == "x"


def test_platform_factor_filter_for_env() -> None:
    value = dedent(
        """
        linux: Linux command
        darwin: Darwin command
        win32: Windows command
        default command
        """,
    )

    result = filter_for_env(value, name="task")
    assert "default command" in result
    if sys.platform == "linux":
        assert "Linux command" in result
        assert "Darwin command" not in result
        assert "Windows command" not in result
    elif sys.platform == "darwin":
        assert "Darwin command" in result
        assert "Linux command" not in result
        assert "Windows command" not in result
    elif sys.platform == "win32":
        assert "Windows command" in result
        assert "Linux command" not in result
        assert "Darwin command" not in result


def test_platform_factor(tox_ini_conf: ToxIniCreator) -> None:
    config = tox_ini_conf(
        """
        [testenv:task]
        commands =
            linux: python -c 'print("Linux")'
            darwin: python -c 'print("Darwin")'
            win32: python -c 'print("Windows")'
            default command
        """,
    )
    env_config = config.get_env("task")
    env_config.add_config(keys="commands", of_type=list[str], default=[], desc="commands")
    commands = env_config["commands"]

    assert "default command" in str(commands)
    if sys.platform == "linux":
        assert 'print("Linux")' in str(commands)
        assert 'print("Darwin")' not in str(commands)
        assert 'print("Windows")' not in str(commands)
    elif sys.platform == "darwin":
        assert 'print("Darwin")' in str(commands)
        assert 'print("Linux")' not in str(commands)
        assert 'print("Windows")' not in str(commands)
    elif sys.platform == "win32":
        assert 'print("Windows")' in str(commands)
        assert 'print("Linux")' not in str(commands)
        assert 'print("Darwin")' not in str(commands)



from __future__ import annotations

import sys
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from tox.tox_env.python.virtual_env.subprocess_adapter import (
    SubprocessCreator,
    SubprocessPythonInfo,
    SubprocessSession,
    _bootstrap_path,  # noqa: PLC2701
    _has_correct_virtualenv,  # noqa: PLC2701
    _VersionInfo,  # noqa: PLC2701
    ensure_bootstrap,
    probe_python,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tox.pytest import ToxProjectCreator


def test_probe_python_current_interpreter() -> None:
    info = probe_python(sys.executable)
    assert info is not None
    assert info.version_info.major == sys.version_info.major
    assert info.version_info.minor == sys.version_info.minor
    assert info.version_info.micro == sys.version_info.micro
    assert info.implementation == sys.implementation.name
    assert info.architecture in {32, 64}
    assert info.system_executable
    assert info.sysconfig_platform is not None
    assert info.machine  # should be non-empty on all platforms


def test_probe_python_nonexistent() -> None:
    assert probe_python("/nonexistent/python999") is None


@pytest.mark.parametrize(
    ("sysconfig_platform", "expected_machine"),
    [
        ("linux-x86_64", "x86_64"),
        ("linux-aarch64", "aarch64"),
        ("win-amd64", "amd64"),
        ("macosx-14.0-arm64", "arm64"),
        ("win32", ""),
        (None, ""),
    ],
)
def test_subprocess_python_info_machine(sysconfig_platform: str | None, expected_machine: str) -> None:
    vi = _VersionInfo(major=3, minor=12, micro=0, releaselevel="final", serial=0)
    info = SubprocessPythonInfo(
        implementation="cpython",
        version_info=vi,
        version="3.12.0",
        architecture=64,
        platform="linux",
        system_executable="/usr/bin/python3",
        free_threaded=False,
        sysconfig_platform=sysconfig_platform,
    )
    assert info.machine == expected_machine


def test_probe_python_failing_script(tmp_path: Path) -> None:
    if sys.platform == "win32":
        bad_script = tmp_path / "bad_python.cmd"
        bad_script.write_text("@exit /b 1\n")
    else:
        bad_script = tmp_path / "bad_python"
        bad_script.write_text("#!/bin/sh\nexit 1\n")
        bad_script.chmod(0o755)
    assert probe_python(str(bad_script)) is None


def test_subprocess_creator_paths_unix(tmp_path: Path) -> None:
    vi = _VersionInfo(major=3, minor=11, micro=0, releaselevel="final", serial=0)
    env_dir = tmp_path / "env"
    info = SubprocessPythonInfo(
        implementation="cpython",
        version_info=vi,
        version="3.11.0",
        architecture=64,
        platform="linux",
        system_executable="/usr/bin/python3.11",
        free_threaded=False,
    )
    creator = SubprocessCreator(_env_dir=env_dir, interpreter=info, _is_win=False)
    assert creator.bin_dir == env_dir / "bin"
    assert creator.script_dir == env_dir / "bin"
    assert creator.purelib == env_dir / "lib" / "python3.11" / "site-packages"
    assert creator.platlib == env_dir / "lib" / "python3.11" / "site-packages"
    assert creator.exe == env_dir / "bin" / "python"


def test_subprocess_creator_paths_windows(tmp_path: Path) -> None:
    vi = _VersionInfo(major=3, minor=11, micro=0, releaselevel="final", serial=0)
    env_dir = tmp_path / "env"
    info = SubprocessPythonInfo(
        implementation="cpython",
        version_info=vi,
        version="3.11.0",
        architecture=64,
        platform="win32",
        system_executable="C:\\Python311\\python.exe",
        free_threaded=False,
    )
    creator = SubprocessCreator(_env_dir=env_dir, interpreter=info, _is_win=True)
    assert creator.bin_dir == env_dir / "Scripts"
    assert creator.purelib == env_dir / "Lib" / "site-packages"
    assert creator.exe == env_dir / "Scripts" / "python.exe"


def test_subprocess_session_run_not_found(tmp_path: Path) -> None:
    vi = _VersionInfo(major=3, minor=11, micro=0, releaselevel="final", serial=0)
    info = SubprocessPythonInfo(
        implementation="cpython",
        version_info=vi,
        version="3.11.0",
        architecture=64,
        platform=sys.platform,
        system_executable=sys.executable,
        free_threaded=False,
    )
    session = SubprocessSession(
        env_dir=tmp_path / "env",
        bootstrap_python=Path("/nonexistent/python"),
        env_vars={},
        interpreter=info,
    )
    with pytest.raises(RuntimeError, match="virtualenv subprocess failed"):
        session.run()


def test_subprocess_session_run_nonzero_exit(tmp_path: Path, mocker: MockerFixture) -> None:
    vi = _VersionInfo(major=3, minor=11, micro=0, releaselevel="final", serial=0)
    info = SubprocessPythonInfo(
        implementation="cpython",
        version_info=vi,
        version="3.11.0",
        architecture=64,
        platform=sys.platform,
        system_executable=sys.executable,
        free_threaded=False,
    )
    session = SubprocessSession(
        env_dir=tmp_path / "env",
        bootstrap_python=Path(sys.executable),
        env_vars={},
        interpreter=info,
    )
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run",
        return_value=mocker.MagicMock(returncode=1, stderr="error"),
    )
    with pytest.raises(RuntimeError, match="virtualenv subprocess failed \\(exit 1\\)"):
        session.run()


def test_subprocess_session_no_interpreter(tmp_path: Path) -> None:
    session = SubprocessSession(
        env_dir=tmp_path / "env",
        bootstrap_python=tmp_path / "bootstrap" / "bin" / "python",
        env_vars={},
        interpreter=None,
    )
    with pytest.raises(RuntimeError, match="no interpreter discovered"):
        _ = session.creator


def test_bootstrap_path_deterministic(tmp_path: Path) -> None:
    path1 = _bootstrap_path(tmp_path, "virtualenv<20.22.0")
    path2 = _bootstrap_path(tmp_path, "virtualenv<20.22.0")
    path3 = _bootstrap_path(tmp_path, "virtualenv<21.0.0")
    assert path1 == path2
    assert path1 != path3
    assert path1.parent == tmp_path / ".virtualenv-bootstrap"


def test_has_correct_virtualenv_nonexistent(tmp_path: Path) -> None:
    assert _has_correct_virtualenv(tmp_path / "nonexistent", "virtualenv") is False


def test_has_correct_virtualenv_matching(tmp_path: Path, mocker: MockerFixture) -> None:
    python = tmp_path / "python"
    python.touch()
    run_mock = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run")
    run_mock.return_value = mocker.MagicMock(returncode=0, stdout="20.21.1\n")
    assert _has_correct_virtualenv(python, "virtualenv<20.22.0") is True


def test_has_correct_virtualenv_bare_spec(tmp_path: Path, mocker: MockerFixture) -> None:
    python = tmp_path / "python"
    python.touch()
    run_mock = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run")
    run_mock.return_value = mocker.MagicMock(returncode=0, stdout="20.21.1\n")
    assert _has_correct_virtualenv(python, "virtualenv") is True


def test_has_correct_virtualenv_subprocess_fails(tmp_path: Path, mocker: MockerFixture) -> None:
    python = tmp_path / "python"
    python.touch()
    run_mock = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run")
    run_mock.return_value = mocker.MagicMock(returncode=1)
    assert _has_correct_virtualenv(python, "virtualenv<20.22.0") is False


def test_has_correct_virtualenv_subprocess_error(tmp_path: Path, mocker: MockerFixture) -> None:
    python = tmp_path / "python"
    python.touch()
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run",
        side_effect=OSError("boom"),
    )
    assert _has_correct_virtualenv(python, "virtualenv<20.22.0") is False


def test_ensure_bootstrap_creates_and_caches(tmp_path: Path, mocker: MockerFixture) -> None:
    venv_create = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.venv.create")
    run_mock = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run")
    run_mock.return_value = mocker.MagicMock(returncode=0, stdout="20.21.1\n")

    has_correct = mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter._has_correct_virtualenv",
        side_effect=[False, False, True],
    )

    result = ensure_bootstrap(tmp_path, "virtualenv<20.22.0")
    assert result.name == ("python.exe" if sys.platform == "win32" else "python")
    venv_create.assert_called_once()
    run_mock.assert_called_once()

    result2 = ensure_bootstrap(tmp_path, "virtualenv<20.22.0")
    assert result == result2
    assert has_correct.call_count == 3


def test_ensure_bootstrap_race_inside_lock(tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter._has_correct_virtualenv",
        side_effect=[False, True],
    )
    result = ensure_bootstrap(tmp_path, "virtualenv<20.22.0")
    assert result.name == ("python.exe" if sys.platform == "win32" else "python")


def test_ensure_bootstrap_removes_stale_base(tmp_path: Path, mocker: MockerFixture) -> None:
    base = _bootstrap_path(tmp_path, "virtualenv<20.22.0")
    base.mkdir(parents=True)
    (base / "stale_file").touch()

    mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.venv.create")
    run_mock = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run")
    run_mock.return_value = mocker.MagicMock(returncode=1, stderr="fail")
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter._has_correct_virtualenv",
        return_value=False,
    )

    with pytest.raises(RuntimeError, match="failed to install"):
        ensure_bootstrap(tmp_path, "virtualenv<20.22.0")
    assert not (base / "stale_file").exists()


def test_ensure_bootstrap_install_failure(tmp_path: Path, mocker: MockerFixture) -> None:
    mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.venv.create")
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter._has_correct_virtualenv",
        return_value=False,
    )
    run_mock = mocker.patch("tox.tox_env.python.virtual_env.subprocess_adapter.subprocess.run")
    run_mock.return_value = mocker.MagicMock(returncode=1, stderr="no matching distribution")

    with pytest.raises(RuntimeError, match="failed to install"):
        ensure_bootstrap(tmp_path, "virtualenv==999.999.999")


def test_virtualenv_spec_config_shown(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_run_base]
            package = "skip"
            virtualenv_spec = "virtualenv<20.22.0"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("c", "-e", "py", "-k", "virtualenv_spec")
    result.assert_success()
    assert "virtualenv<20.22.0" in result.out


def test_virtualenv_spec_cache_includes_spec(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter.ensure_bootstrap",
        return_value=Path(sys.executable),
    )
    mocker.patch(
        "tox.tox_env.python.virtual_env.subprocess_adapter.probe_python",
        return_value=SubprocessPythonInfo(
            implementation=sys.implementation.name,
            version_info=_VersionInfo(*sys.version_info[:5]),
            version=f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            architecture=64,
            platform=sys.platform,
            system_executable=sys.executable,
            free_threaded=False,
        ),
    )

    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_run_base]
            package = "skip"
            virtualenv_spec = "virtualenv<20.22.0"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()


def test_virtualenv_spec_empty_uses_imported(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_run_base]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()
    assert "ok" in result.out



from __future__ import annotations

from typing import TYPE_CHECKING

from tox.util.path import ensure_cachedir_tag, ensure_empty_dir, ensure_gitignore

if TYPE_CHECKING:
    from pathlib import Path

_EXPECTED_CACHEDIR_TAG = """\
Signature: 8a477f597d28d172789f06886806bc55
# This file is a cache directory tag created by tox.
# For information about cache directory tags, see:
#\thttps://bford.info/cachedir/spec.html
"""


def test_ensure_cachedir_tag_creates_file(tmp_path: Path) -> None:
    ensure_cachedir_tag(tmp_path)
    tag = tmp_path / "CACHEDIR.TAG"
    assert tag.is_file()
    content = tag.read_text(encoding="utf-8")
    assert content.startswith("Signature: 8a477f597d28d172789f06886806bc55\n")
    assert content == _EXPECTED_CACHEDIR_TAG


def test_ensure_cachedir_tag_creates_parent_dirs(tmp_path: Path) -> None:
    nested = tmp_path / "a" / "b"
    ensure_cachedir_tag(nested)
    assert (nested / "CACHEDIR.TAG").is_file()


def test_ensure_cachedir_tag_idempotent(tmp_path: Path) -> None:
    ensure_cachedir_tag(tmp_path)
    tag = tmp_path / "CACHEDIR.TAG"
    first_content = tag.read_text(encoding="utf-8")
    ensure_cachedir_tag(tmp_path)
    assert tag.read_text(encoding="utf-8") == first_content


def test_ensure_cachedir_tag_does_not_overwrite(tmp_path: Path) -> None:
    tag = tmp_path / "CACHEDIR.TAG"
    tag.write_text("Signature: 8a477f597d28d172789f06886806bc55\n# custom\n", encoding="utf-8")
    ensure_cachedir_tag(tmp_path)
    assert tag.read_text(encoding="utf-8") == "Signature: 8a477f597d28d172789f06886806bc55\n# custom\n"


def test_ensure_empty_dir_file(tmp_path: Path) -> None:
    dest = tmp_path / "a"
    dest.write_text("")
    ensure_empty_dir(dest)
    assert dest.is_dir()
    assert not list(dest.iterdir())


def test_ensure_gitignore_creates_file(tmp_path: Path) -> None:
    target = tmp_path / "work"
    target.mkdir()
    ensure_gitignore(target)
    gitignore = target / ".gitignore"
    assert gitignore.exists()
    assert gitignore.read_text(encoding="utf-8") == "*\n"


def test_ensure_gitignore_creates_parent_dirs(tmp_path: Path) -> None:
    target = tmp_path / "a" / "b" / "c"
    ensure_gitignore(target)
    gitignore = target / ".gitignore"
    assert gitignore.exists()
    assert gitignore.read_text(encoding="utf-8") == "*\n"


def test_ensure_gitignore_does_not_overwrite(tmp_path: Path) -> None:
    target = tmp_path / "work"
    target.mkdir()
    gitignore = target / ".gitignore"
    gitignore.write_text("custom\n", encoding="utf-8")
    ensure_gitignore(target)
    assert gitignore.read_text(encoding="utf-8") == "custom\n"


def test_ensure_gitignore_idempotent(tmp_path: Path) -> None:
    target = tmp_path / "work"
    target.mkdir()
    ensure_gitignore(target)
    ensure_gitignore(target)
    assert (target / ".gitignore").read_text(encoding="utf-8") == "*\n"



"""Generate schema for tox configuration, respecting the current plugins."""

from __future__ import annotations

import json
import sys
import typing
from pathlib import Path
from types import UnionType
from typing import TYPE_CHECKING

import packaging.requirements
import packaging.version

import tox.config.set_env
import tox.config.types
import tox.tox_env.python.pip.req_file
from tox.config.cli.parser import CORE
from tox.plugin import impl

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.config.sets import ConfigSet
    from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command(
        "schema", [], "Generate schema for tox configuration", gen_schema, inherit=frozenset({CORE})
    )
    our.add_argument("--strict", action="store_true", help="Disallow extra properties in configuration")


def gen_schema(state: State) -> int:
    core = state.conf.core
    strict = state.conf.options.strict

    # Use any available run environment for introspection (fall back to "py" which is always defined)
    env_name = next(state.envs.iter(only_active=False), "py")
    env_properties = _get_schema(state.envs[env_name].conf, path="#/properties/env_run_base/properties")

    properties = _get_schema(core, path="#/properties")

    # This accesses plugins that register new sections (like tox-gh)
    # Accessing a private member since this is not exposed yet and the
    # interface includes the internal storage tuple
    sections = {
        key: conf
        for s, conf in state.conf._key_to_conf_set.items()  # noqa: SLF001
        if (key := s[0].split(".")[0]) not in {"env_run_base", "env_pkg_base", "env"}
    }
    for key, conf in sections.items():
        properties[key] = {
            "type": "object",
            "additionalProperties": not strict,
            "properties": _get_schema(conf, path=f"#/properties/{key}/properties"),
        }

    docs_base = "https://tox.wiki/en/stable"
    json_schema = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "https://raw.githubusercontent.com/tox-dev/tox/main/src/tox/tox.schema.json",
        "title": "tox configuration",
        "description": "tox configuration file (tox.toml or [tool.tox] in pyproject.toml)",
        "x-taplo": {"links": {"key": f"{docs_base}/config.html"}},
        "type": "object",
        "properties": {
            **properties,
            "env_run_base": {
                "type": "object",
                "description": "base configuration for run environments",
                "x-taplo": {"links": {"key": f"{docs_base}/config.html#run-environment"}},
                "properties": env_properties,
                "additionalProperties": not strict,
            },
            "env_pkg_base": {
                "type": "object",
                "$ref": "#/properties/env_run_base",
                "description": "base configuration for packaging environments",
                "x-taplo": {"links": {"key": f"{docs_base}/config.html#packaging-environment"}},
                "additionalProperties": not strict,
            },
            "env": {
                "type": "object",
                "description": "per-environment overrides (keyed by environment name)",
                "x-taplo": {"links": {"key": f"{docs_base}/config.html#run-environment"}},
                "patternProperties": {"^.*$": {"$ref": "#/properties/env_run_base"}},
            },
            "legacy_tox_ini": {
                "type": "string",
                "description": "tox configuration in INI format embedded in a TOML file",
                "x-taplo": {"links": {"key": f"{docs_base}/config.html#pyproject-toml-ini"}},
            },
        },
        "additionalProperties": not strict,
        "definitions": {
            "subs": {
                "anyOf": [
                    {"type": "string"},
                    {
                        "type": "object",
                        "properties": {
                            "replace": {"type": "string"},
                            "name": {"type": "string"},
                            "default": {
                                "oneOf": [
                                    {"type": "string"},
                                    {"type": "array", "items": {"$ref": "#/definitions/subs"}},
                                ]
                            },
                            "extend": {"type": "boolean"},
                        },
                        "required": ["replace"],
                        "additionalProperties": False,
                    },
                    {
                        "type": "object",
                        "properties": {
                            "replace": {"type": "string"},
                            "of": {"type": "array", "items": {"type": "string"}},
                            "default": {
                                "oneOf": [
                                    {"type": "string"},
                                    {"type": "array", "items": {"$ref": "#/definitions/subs"}},
                                ]
                            },
                            "extend": {"type": "boolean"},
                        },
                        "required": ["replace", "of"],
                        "additionalProperties": False,
                    },
                ],
            },
        },
    }
    print(json.dumps(json_schema, indent=2))  # noqa: T201
    return 0


def _get_schema(conf: ConfigSet, path: str) -> dict[str, dict[str, typing.Any]]:
    properties: dict[str, dict[str, typing.Any]] = {}
    for x in conf.get_configs():
        name, *aliases = x.keys
        if (of_type := getattr(x, "of_type", None)) is None:
            continue
        desc = getattr(x, "desc", None)
        try:
            properties[name] = {**_process_type(of_type), "description": desc}
        except ValueError:
            print(name, "has unrecoginsed type:", of_type, file=sys.stderr)  # noqa: T201
        for alias in aliases:
            properties[alias] = {
                "$ref": f"{path}/{name}",
                "description": f"Deprecated: use {name!r} instead",
                "deprecated": True,
            }
    return properties


def _process_type(of_type: typing.Any) -> dict[str, typing.Any]:  # noqa: C901, PLR0911, PLR0912
    if of_type in {
        Path,
        str,
        packaging.version.Version,
        packaging.requirements.Requirement,
        tox.tox_env.python.pip.req_file.PythonDeps,
        tox.tox_env.python.pip.req_file.PythonConstraints,
    }:
        return {"type": "string"}
    if typing.get_origin(of_type) is typing.Union or isinstance(of_type, UnionType):
        types = [x for x in typing.get_args(of_type) if x is not type(None)]
        if len(types) == 1:
            return _process_type(types[0])
        return {"oneOf": [_process_type(t) for t in types]}
    if of_type is bool:
        return {"type": "boolean"}
    if of_type is int:
        return {"type": "integer", "minimum": 0}
    if of_type is float:
        return {"type": "number"}
    if typing.get_origin(of_type) is typing.Literal:
        return {"enum": list(typing.get_args(of_type))}
    if of_type is tox.config.types.EnvList:
        return {
            "type": "array",
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/subs"},
                    {
                        "type": "object",
                        "required": ["product"],
                        "properties": {
                            "product": {
                                "type": "array",
                                "items": {
                                    "oneOf": [
                                        {"type": "array", "items": {"type": "string"}},
                                        {
                                            "type": "object",
                                            "required": ["prefix"],
                                            "properties": {
                                                "prefix": {"type": "string"},
                                                "start": {"type": "integer"},
                                                "stop": {"type": "integer"},
                                            },
                                            "additionalProperties": False,
                                        },
                                        {
                                            "type": "object",
                                            "minProperties": 1,
                                            "maxProperties": 1,
                                            "not": {"required": ["prefix"]},
                                            "additionalProperties": {
                                                "type": "array",
                                                "items": {"type": "string"},
                                            },
                                            "description": "labeled factor group for {factor:label} substitution",
                                        },
                                    ],
                                },
                                "description": "factor groups for cartesian product expansion",
                            },
                            "exclude": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "environment names to exclude from product",
                            },
                        },
                        "additionalProperties": False,
                    },
                ],
            },
        }
    if of_type is tox.config.types.Command:
        return {"type": "array", "items": {"$ref": "#/definitions/subs"}}
    if typing.get_origin(of_type) in {list, set}:
        if typing.get_args(of_type)[0] in {str, packaging.requirements.Requirement}:
            return {"type": "array", "items": {"$ref": "#/definitions/subs"}}
        if typing.get_args(of_type)[0] is tox.config.types.Command:
            return {"type": "array", "items": _process_type(typing.get_args(of_type)[0])}
        msg = f"Unknown list type: {of_type}"
        raise ValueError(msg)
    if of_type is tox.config.set_env.SetEnv:
        return {
            "type": "object",
            "additionalProperties": {"$ref": "#/definitions/subs"},
        }
    if typing.get_origin(of_type) is dict:
        return {
            "type": "object",
            "additionalProperties": {**_process_type(typing.get_args(of_type)[1])},
        }
    msg = f"Unknown type: {of_type}"
    raise ValueError(msg)



from __future__ import annotations

import logging
import operator
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Sequence
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from packaging.requirements import Requirement

from tox.config.types import Command
from tox.execute.request import StdinSource
from tox.tox_env.errors import Fail, Recreate
from tox.tox_env.installer import Installer
from tox.tox_env.python.api import Python
from tox.tox_env.python.package import EditableLegacyPackage, EditablePackage, SdistPackage, WheelPackage
from tox.tox_env.python.pip.req_file import PythonConstraints, PythonDeps
from tox.tox_env.python.pylock import Pylock

if TYPE_CHECKING:
    from tox.config.main import Config
    from tox.tox_env.package import PathPackage


class PythonInstallerListDependencies(Installer[Python], ABC):
    def __init__(self, tox_env: Python, with_list_deps: bool = True) -> None:  # noqa: FBT001, FBT002
        self._with_list_deps = with_list_deps
        super().__init__(tox_env)

    def _register_config(self) -> None:
        if self._with_list_deps:  # pragma: no branch
            self._env.conf.add_config(
                keys=["list_dependencies_command"],
                of_type=Command,
                default=Command(self.freeze_cmd()),
                desc="command used to list installed packages",
            )

    @abstractmethod
    def freeze_cmd(self) -> list[str]:
        raise NotImplementedError

    def installed(self) -> list[str]:
        cmd: Command = self._env.conf["list_dependencies_command"]
        result = self._env.execute(cmd=cmd.args, stdin=StdinSource.OFF, run_id="freeze", show=False)
        result.assert_success()
        return result.out.splitlines()


class Pip(PythonInstallerListDependencies):
    """Pip is a python installer that can install packages as defined by PEP-508 and PEP-517."""

    def _register_config(self) -> None:
        super()._register_config()
        root = self._env.core["toxinidir"]
        self._env.conf.add_config(
            keys=["pip_pre"],
            of_type=bool,
            default=False,
            desc="install the latest available pre-release (alpha/beta/rc) of dependencies without a specified version",
        )
        self._env.conf.add_config(  # ty: ignore[no-matching-overload] # https://github.com/astral-sh/ty/issues/2428
            keys=["install_command"],
            of_type=Command,
            default=self.default_install_command,
            post_process=self.post_process_install_command,
            desc="command used to install packages",
        )
        self._env.conf.add_config(
            keys=["constraints"],
            of_type=PythonConstraints,
            factory=partial(PythonConstraints.factory, root),
            default=PythonConstraints("", root),
            desc="constraints to apply to installed python dependencies",
        )
        self._env.conf.add_config(
            keys=["constrain_package_deps"],
            of_type=bool,
            default=False,
            desc="If true, apply constraints during install_package_deps.",
        )
        self._env.conf.add_config(
            keys=["use_frozen_constraints"],
            of_type=bool,
            default=False,
            desc="Use the exact versions of installed deps as constraints, otherwise use the listed deps.",
        )

    def freeze_cmd(self) -> list[str]:  # noqa: PLR6301
        return ["python", "-m", "pip", "freeze", "--all"]

    def default_install_command(self, conf: Config, env_name: str | None) -> Command:  # noqa: ARG002
        isolated_flag = "-E" if self._env.base_python.version_info.major == 2 else "-I"  # noqa: PLR2004
        cmd = Command(["python", isolated_flag, "-m", "pip", "install", "{opts}", "{packages}"])
        return self.post_process_install_command(cmd)

    def post_process_install_command(self, cmd: Command) -> Command:
        install_command = cmd.args
        pip_pre: bool = self._env.conf["pip_pre"]
        try:
            opts_at = install_command.index("{opts}")
        except ValueError:
            if pip_pre:
                install_command.append("--pre")
        else:
            if pip_pre:
                install_command[opts_at] = "--pre"
            else:
                install_command.pop(opts_at)
        return cmd

    def install(self, arguments: Any, section: str, of_type: str) -> None:
        if isinstance(arguments, PythonDeps):
            self._install_requirement_file(arguments, section, of_type)
        elif isinstance(arguments, Pylock):
            self._install_pylock(arguments, section, of_type)
        elif isinstance(arguments, Sequence):
            self._install_list_of_deps(arguments, section, of_type)
        else:
            logging.warning("pip cannot install %r", arguments)
            raise SystemExit(1)

    @property
    def constraints(self) -> PythonConstraints:
        return cast("PythonConstraints", self._env.conf["constraints"])

    def constraints_file(self) -> Path:
        return Path(self._env.env_dir) / "constraints.txt"

    @property
    def constrain_package_deps(self) -> bool:
        return bool(self._env.conf["constrain_package_deps"])

    @property
    def use_frozen_constraints(self) -> bool:
        return bool(self._env.conf["use_frozen_constraints"])

    def _install_requirement_file(self, arguments: PythonDeps, section: str, of_type: str) -> None:
        new_requirements: list[str] = []
        new_constraints: list[str] = []

        try:
            new_options, new_reqs = arguments.unroll()
        except ValueError as exception:
            msg = f"{exception} for tox env py within deps"
            raise Fail(msg) from exception
        for req in new_reqs:
            (new_constraints if req.startswith("-c ") else new_requirements).append(req)

        try:
            _, new_reqs = self.constraints.unroll()
        except ValueError as exception:
            msg = f"{exception} for tox env py within constraints"
            raise Fail(msg) from exception
        new_constraints.extend(new_reqs)

        constraint_options = {
            "constrain_package_deps": self.constrain_package_deps,
            "use_frozen_constraints": self.use_frozen_constraints,
        }
        new = {
            "options": new_options,
            "requirements": new_requirements,
            "constraints": new_constraints,
            "constraint_options": constraint_options,
        }
        # if option or constraint change in any way recreate, if the requirements change only if some are removed
        with self._env.cache.compare(new, section, of_type) as (eq, old):
            if not eq:  # pragma: no branch
                if old is not None:
                    self._recreate_if_diff("install flag(s)", new_options, old["options"], lambda i: i)
                    self._recreate_if_diff(
                        "constraint(s)", new_constraints, old["constraints"], operator.itemgetter(slice(3, None))
                    )
                    missing_requirement = set(old["requirements"]) - set(new_requirements)
                    if missing_requirement:
                        msg = f"requirements removed: {' '.join(missing_requirement)}"
                        raise Recreate(msg)
                    old_constraint_options = old.get("constraint_options")
                    if old_constraint_options != constraint_options:
                        msg = f"constraint options changed: old={old_constraint_options} new={constraint_options}"
                        raise Recreate(msg)
                args = arguments.as_root_args
                if args:  # pragma: no branch
                    args.extend(self.constraints.as_root_args)
                    self._execute_installer(args, of_type)
                    if self.constrain_package_deps and not self.use_frozen_constraints:
                        combined_constraints = new_requirements + [c.removeprefix("-c ") for c in new_constraints]
                        self.constraints_file().write_text("\n".join(combined_constraints))

    @staticmethod
    def _recreate_if_diff(of_type: str, new_opts: list[str], old_opts: list[str], fmt: Callable[[str], str]) -> None:
        if old_opts == new_opts:
            return
        removed_opts = set(old_opts) - set(new_opts)
        removed = f" removed {', '.join(sorted(fmt(i) for i in removed_opts))}" if removed_opts else ""
        added_opts = set(new_opts) - set(old_opts)
        added = f" added {', '.join(sorted(fmt(i) for i in added_opts))}" if added_opts else ""
        msg = f"changed {of_type}{removed}{added}"
        raise Recreate(msg)

    def _install_pylock(self, pylock: Pylock, section: str, of_type: str) -> None:
        requirements = pylock.requirements()
        new_reqs = sorted(str(r) for r in requirements)
        with self._env.cache.compare(new_reqs, section, of_type) as (eq, old):
            if not eq:
                if old is not None and (missing := sorted(set(old) - set(new_reqs))):
                    msg = f"pylock dependencies removed: {' '.join(missing)}"
                    raise Recreate(msg)
                if new_deps := sorted(set(new_reqs) - set(old or [])):
                    req_file = Path(self._env.env_dir) / "pylock.txt"
                    req_file.write_text("\n".join(new_deps))
                    self._execute_installer(["--no-deps", "-r", str(req_file)], of_type)

    def _install_list_of_deps(  # noqa: C901, PLR0912
        self,
        arguments: Sequence[
            Requirement | WheelPackage | SdistPackage | EditableLegacyPackage | EditablePackage | PathPackage
        ],
        section: str,
        of_type: str,
    ) -> None:
        groups: dict[str, list[str]] = defaultdict(list)
        config_settings: dict[str, str] = {}
        for arg in arguments:
            if isinstance(arg, Requirement):
                groups["req"].append(str(arg))
            elif isinstance(arg, (WheelPackage, SdistPackage, EditablePackage)):
                groups["req"].extend(self._apply_force_deps(arg.deps))
                groups["pkg"].append(str(arg.path))
                if isinstance(arg, SdistPackage) and arg.config_settings:
                    config_settings.update(arg.config_settings)
            elif isinstance(arg, EditableLegacyPackage):
                groups["req"].extend(self._apply_force_deps(arg.deps))
                groups["dev_pkg"].append(str(arg.path))
            else:
                logging.warning("pip cannot install %r", arg)
                raise SystemExit(1)
        req_of_type = f"{of_type}_deps" if groups["pkg"] or groups["dev_pkg"] else of_type
        for value in groups.values():
            value.sort()
        with self._env.cache.compare(groups["req"], section, req_of_type) as (eq, old):
            if not eq:  # pragma: no branch
                miss = sorted(set(old or []) - set(groups["req"]))
                if miss:  # no way yet to know what to uninstall here (transitive dependencies?)
                    msg = f"dependencies removed: {', '.join(str(i) for i in miss)}"
                    raise Recreate(msg)  # pragma: no branch
                new_deps = sorted(set(groups["req"]) - set(old or []))
                if new_deps:  # pragma: no branch
                    new_deps.extend(self.constraints.as_root_args)
                    self._execute_installer(new_deps, req_of_type)
        install_args = ["--force-reinstall", "--no-deps"]
        cs_args = [f"--config-settings={k}={v}" for k, v in config_settings.items()]
        if groups["pkg"]:
            # we intentionally ignore constraints when installing the package itself
            # https://github.com/tox-dev/tox/issues/3550
            self._execute_installer(install_args + cs_args + groups["pkg"], of_type)
        if groups["dev_pkg"]:
            for entry in groups["dev_pkg"]:
                install_args.extend(("-e", str(entry)))
            # we intentionally ignore constraints when installing the package itself
            # https://github.com/tox-dev/tox/issues/3550
            self._execute_installer(install_args, of_type)

    def _apply_force_deps(self, deps: Sequence[Requirement]) -> list[str]:
        forced: dict[str, Requirement] = {r.name: r for r in getattr(self._env.options, "force_dep", [])}
        return [str(forced.get(dep.name, dep)) for dep in deps]

    def _execute_installer(self, deps: Sequence[Any], of_type: str) -> None:
        if of_type == "package_deps" and self.constrain_package_deps:
            constraints_file = self.constraints_file()
            if constraints_file.exists():
                deps = [*deps, f"-c{constraints_file}"]

        cmd = self.build_install_cmd(deps)
        outcome = self._env.execute(cmd, stdin=StdinSource.OFF, run_id=f"install_{of_type}")
        outcome.assert_success()

        if of_type == "deps" and self.constrain_package_deps and self.use_frozen_constraints:
            # freeze installed deps for use as constraints
            self.constraints_file().write_text("\n".join(self.installed()))

    def build_install_cmd(self, args: Sequence[str]) -> list[str]:
        try:
            cmd: Command = self._env.conf["install_command"]
        except ValueError as exc:
            msg = f"unable to determine pip install command: {exc!s}"
            raise Fail(msg) from exc
        install_command = cmd.args
        try:
            opts_at = install_command.index("{packages}")
        except ValueError:
            opts_at = len(install_command)
        return install_command[:opts_at] + list(args) + install_command[opts_at + 1 :]


__all__ = [
    "Pip",
    "PythonInstallerListDependencies",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


@pytest.mark.parametrize("value", ["1", "0", "", "arbitrary_value"])
def test_ci_passthrough_present(value: str, tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("CI", value)
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands=python -c 'print(0)'\n"})
    execute_calls = prj.patch_execute(lambda _r: 0)
    result = prj.run("r", "-e", "py")
    result.assert_success()
    req = execute_calls.call_args[0][3]
    assert req.env["__TOX_ENVIRONMENT_VARIABLE_ORIGINAL_CI"] == value


def test_ci_passthrough_absent(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("CI", raising=False)
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands=python -c 'print(0)'\n"})
    execute_calls = prj.patch_execute(lambda _r: 0)
    result = prj.run("r", "-e", "py")
    result.assert_success()
    req = execute_calls.call_args[0][3]
    assert "__TOX_ENVIRONMENT_VARIABLE_ORIGINAL_CI" not in req.env



from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal
from unittest.mock import ANY

import pytest

from tox.config.set_env import SetEnv

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tox.pytest import MonkeyPatch, ToxProjectCreator

from typing import Protocol


def test_set_env_explicit() -> None:
    set_env = SetEnv("\nA=1\nB = 2\nC= 3\nD= 4", "py", "py", Path())
    set_env.update({"E": "5 ", "F": "6"}, override=False)

    keys = list(set_env)
    assert keys == ["E", "F", "A", "B", "C", "D"]
    values = [set_env.load(k) for k in keys]
    assert values == ["5 ", "6", "1", "2", "3", "4"]

    for key in keys:
        assert key in set_env
    assert "MISS" not in set_env


def test_set_env_merge() -> None:
    a = SetEnv("\nA=1\nB = 2\nC= 3\nD= 4", "py", "py", Path())
    b = SetEnv("\nA=2\nE = 5", "py", "py", Path())
    a.update(b, override=False)

    keys = list(a)
    assert keys == ["E", "A", "B", "C", "D"]
    values = [a.load(k) for k in keys]
    assert values == ["5", "1", "2", "3", "4"]

    a.update(b, override=True)

    values = [a.load(k) for k in keys]
    assert values == ["5", "2", "2", "3", "4"]


def test_set_env_bad_line() -> None:
    with pytest.raises(ValueError, match="A"):
        SetEnv("A", "py", "py", Path())


ConfigFileFormat = Literal["ini", "toml"]


class EvalSetEnv(Protocol):
    def __call__(
        self,
        config: str,
        *,
        of_type: ConfigFileFormat = "ini",
        extra_files: dict[str, Any] | None = ...,
        from_cwd: Path | None = ...,
    ) -> SetEnv: ...


@pytest.fixture
def eval_set_env(tox_project: ToxProjectCreator) -> EvalSetEnv:
    def func(
        config: str,
        *,
        of_type: ConfigFileFormat = "ini",
        extra_files: dict[str, Any] | None = None,
        from_cwd: Path | None = None,
    ) -> SetEnv:
        prj = tox_project({f"tox.{of_type}": config, **(extra_files or {})})
        result = prj.run("c", "-k", "set_env", "-e", "py", from_cwd=None if from_cwd is None else prj.path / from_cwd)
        result.assert_success()
        set_env: SetEnv = result.env_conf("py")["set_env"]
        return set_env

    return func


def test_set_env_default(eval_set_env: EvalSetEnv) -> None:
    set_env = eval_set_env("")
    keys = list(set_env)
    assert keys == ["PYTHONHASHSEED", "PIP_DISABLE_PIP_VERSION_CHECK", "PYTHONIOENCODING"]
    values = [set_env.load(k) for k in keys]
    assert values == [ANY, "1", "utf-8"]


def test_set_env_self_key(eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("a", "1")
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env=a={env:a:2}")
    assert set_env.load("a") == "1"


def test_set_env_other_env_set(eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("b", "1")
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env=a={env:b:2}")
    assert set_env.load("a") == "1"


def test_set_env_other_env_default(eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.delenv("b", raising=False)
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env=a={env:b:2}")
    assert set_env.load("a") == "2"


def test_set_env_delayed_eval(eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("b", "c=1")
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env={env:b}")
    assert set_env.load("c") == "1"


def test_set_env_tty_on(eval_set_env: EvalSetEnv, mocker: MockerFixture) -> None:
    mocker.patch("sys.stdout.isatty", return_value=True)
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env={tty:A=1:B=1}")
    assert set_env.load("A") == "1"
    assert "B" not in set_env


def test_set_env_tty_off(eval_set_env: EvalSetEnv, mocker: MockerFixture) -> None:
    mocker.patch("sys.stdout.isatty", return_value=False)
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env={tty:A=1:B=1}")
    assert set_env.load("B") == "1"
    assert "A" not in set_env


def test_set_env_circular_use_os_environ(tox_project: ToxProjectCreator) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip\nset_env=a={env:b}\n b={env:a}"})
    result = prj.run("c", "-e", "py", raise_on_config_fail=False)
    result.assert_failed(code=-1)
    assert "replace failed in py.set_env with MatchRecursionError" in result.out, result.out
    assert "circular chain between set env a, b" in result.out, result.out


def test_set_env_invalid_lines(eval_set_env: EvalSetEnv) -> None:
    with pytest.raises(ValueError, match="a"):
        eval_set_env("[testenv]\npackage=skip\nset_env=a\n b")


def test_set_env_replacer(eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("MAGIC", "\nb=2\n")
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env=a=1\n {env:MAGIC}")
    env = {k: set_env.load(k) for k in set_env}
    assert env == {
        "PIP_DISABLE_PIP_VERSION_CHECK": "1",
        "a": "1",
        "b": "2",
        "PYTHONIOENCODING": "utf-8",
        "PYTHONHASHSEED": ANY,
    }


def test_set_env_honor_override(eval_set_env: EvalSetEnv) -> None:
    set_env = eval_set_env("[testenv]\npackage=skip\nset_env=PIP_DISABLE_PIP_VERSION_CHECK=0")
    assert set_env.load("PIP_DISABLE_PIP_VERSION_CHECK") == "0"


@pytest.mark.parametrize(
    ("of_type", "config"),
    [
        pytest.param("ini", "[testenv]\npackage=skip\nset_env=file|A{/}a.txt\nchange_dir=C", id="ini"),
        pytest.param("toml", '[env_run_base]\npackage="skip"\nset_env={file="A{/}a.txt"}\nchange_dir="C"', id="toml"),
        pytest.param("ini", "[testenv]\npackage=skip\nset_env=file|{env:env_file}\nchange_dir=C", id="ini-env"),
        pytest.param(
            "toml", '[env_run_base]\npackage="skip"\nset_env={file="{env:env_file}"}\nchange_dir="C"', id="toml-env"
        ),
    ],
)
def test_set_env_environment_file(
    of_type: ConfigFileFormat, config: str, eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch
) -> None:
    env_file = """
    A=1
    B= 2
    C = 1
    # D = comment # noqa: E800
    E = "1"
    F =
    """
    monkeypatch.setenv("env_file", "A{/}a.txt")

    extra = {"A": {"a.txt": env_file}, "B": None, "C": None}
    set_env = eval_set_env(config, of_type=of_type, extra_files=extra, from_cwd=Path("B"))
    content = {k: set_env.load(k) for k in set_env}
    assert content == {
        "PIP_DISABLE_PIP_VERSION_CHECK": "1",
        "PYTHONHASHSEED": ANY,
        "A": "1",
        "B": "2",
        "C": "1",
        "E": '"1"',
        "F": "",
        "PYTHONIOENCODING": "utf-8",
    }


@pytest.mark.parametrize(
    ("of_type", "config"),
    [
        pytest.param("ini", "[testenv]\npackage=skip\nset_env=file|A{/}a.txt\n X=y\nchange_dir=C", id="ini"),
        pytest.param(
            "toml", '[env_run_base]\npackage="skip"\nset_env={file="A{/}a.txt", X="y"}\nchange_dir="C"', id="toml"
        ),
        pytest.param("ini", "[testenv]\npackage=skip\nset_env=file|{env:env_file}\n X=y\nchange_dir=C", id="ini-env"),
        pytest.param(
            "toml",
            '[env_run_base]\npackage="skip"\nset_env={file="{env:env_file}", X="y"}\nchange_dir="C"',
            id="toml-env",
        ),
    ],
)
def test_set_env_environment_file_combined_with_normal_setting(
    of_type: ConfigFileFormat, config: str, eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch
) -> None:
    env_file = """
    A=1
    """
    # Monkeypatch only used for some of the parameters
    monkeypatch.setenv("env_file", "A{/}a.txt")

    extra = {"A": {"a.txt": env_file}, "B": None, "C": None}
    set_env = eval_set_env(config, of_type=of_type, extra_files=extra, from_cwd=Path("B"))
    content = {k: set_env.load(k) for k in set_env}
    assert content == {
        "PIP_DISABLE_PIP_VERSION_CHECK": "1",
        "PYTHONHASHSEED": ANY,
        "A": "1",
        "X": "y",
        "PYTHONIOENCODING": "utf-8",
    }


def test_set_env_file_does_not_override_later_values(tox_project: ToxProjectCreator) -> None:
    ini = """\
    [testenv]
    skip_install = true
    set_env =
        file|.env
        FOO=QUX
    """
    project = tox_project({"tox.ini": ini, ".env": "FOO=BAR\nEXTRA=from_file"})
    result = project.run("c", "-e", "py", "-k", "set_env")
    result.assert_success()
    set_env = result.env_conf("py")["set_env"]
    content = {k: set_env.load(k) for k in set_env}
    assert content["FOO"] == "QUX"
    assert content["EXTRA"] == "from_file"


def test_set_env_environment_file_missing(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\npackage=skip\nset_env=file|magic.txt"})
    result = project.run("r")
    result.assert_failed()
    assert f"py: failed with {project.path / 'magic.txt'} does not exist for set_env" in result.out


# https://github.com/tox-dev/tox/issues/2435
def test_set_env_environment_with_file_and_expanded_substitution(
    tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch
) -> None:
    conf = {
        "tox.ini": """
        [tox]
        envlist =
            check

        [testenv]
        setenv =
            file|.env
            PRECENDENCE_TEST_1=1_expanded_precedence

        [testenv:check]
        setenv =
            {[testenv]setenv}
            PRECENDENCE_TEST_1=1_self_precedence
            PRECENDENCE_TEST_2=2_self_precedence
        """,
        ".env": """
        PRECENDENCE_TEST_1=1_file_precedence
        PRECENDENCE_TEST_2=2_file_precedence
        PRECENDENCE_TEST_3=3_file_precedence
        """,
    }
    monkeypatch.setenv("env_file", ".env")
    project = tox_project(conf)

    result = project.run("c", "-k", "set_env", "-e", "check")
    result.assert_success()
    set_env = result.env_conf("check")["set_env"]
    content = {k: set_env.load(k) for k in set_env}
    assert content == {
        "PIP_DISABLE_PIP_VERSION_CHECK": "1",
        "PYTHONHASHSEED": ANY,
        "PYTHONIOENCODING": "utf-8",
        "PRECENDENCE_TEST_1": "1_self_precedence",
        "PRECENDENCE_TEST_2": "2_self_precedence",
        "PRECENDENCE_TEST_3": "3_file_precedence",
    }

    result = project.run("r", "-e", "check")
    result.assert_success()
    assert "check: OK" in result.out


@pytest.mark.parametrize(
    ("of_type", "config", "expected_present", "expected_value"),
    [
        pytest.param(
            "ini",
            f"[testenv]\npackage=skip\nset_env=CONDITIONAL=value; sys_platform == '{sys.platform}'",
            True,
            "value",
            id="ini-marker-true",
        ),
        pytest.param(
            "ini",
            "[testenv]\npackage=skip\nset_env=CONDITIONAL=value; sys_platform == 'nonexistent'",
            False,
            None,
            id="ini-marker-false",
        ),
        pytest.param(
            "toml",
            f'[env_run_base]\npackage="skip"\nset_env.CONDITIONAL = {{ value = "value", '
            f"marker = \"sys_platform == '{sys.platform}'\" }}",
            True,
            "value",
            id="toml-marker-true",
        ),
        pytest.param(
            "toml",
            '[env_run_base]\npackage="skip"\n'
            'set_env.CONDITIONAL = { value = "value", marker = "sys_platform == \'nonexistent\'" }',
            False,
            None,
            id="toml-marker-false",
        ),
        pytest.param(
            "ini",
            "[testenv]\npackage=skip\nset_env=UNCONDITIONAL=value",
            True,
            "value",
            id="ini-no-marker",
        ),
        pytest.param(
            "ini",
            f"[testenv]\npackage=skip\nset_env=OS_CHECK=yes; os_name == '{os.name}'",
            True,
            "yes",
            id="ini-os-name-marker",
        ),
    ],
)
def test_set_env_marker(
    eval_set_env: EvalSetEnv,
    of_type: ConfigFileFormat,
    config: str,
    expected_present: bool,
    expected_value: str | None,
) -> None:
    set_env = eval_set_env(config, of_type=of_type)
    if expected_present:
        assert "CONDITIONAL" in set_env or "UNCONDITIONAL" in set_env or "OS_CHECK" in set_env
        key = next(k for k in ("CONDITIONAL", "UNCONDITIONAL", "OS_CHECK") if k in set_env)
        assert set_env.load(key) == expected_value
    else:
        assert "CONDITIONAL" not in set_env


@pytest.mark.parametrize(
    ("marker", "expected_present"),
    [
        pytest.param(f"sys_platform == '{sys.platform}'", True, id="marker-true"),
        pytest.param("sys_platform == 'nonexistent'", False, id="marker-false"),
    ],
)
def test_set_env_marker_with_replace_toml(
    eval_set_env: EvalSetEnv, monkeypatch: MonkeyPatch, marker: str, expected_present: bool
) -> None:
    monkeypatch.setenv("MY_VAR", "from_env")
    config = (
        f'[env_run_base]\npackage="skip"\n'
        f'set_env.CONDITIONAL = {{ replace = "env", name = "MY_VAR", default = "default", marker = "{marker}" }}'
    )
    set_env = eval_set_env(config, of_type="toml")
    if expected_present:
        assert "CONDITIONAL" in set_env
        assert set_env.load("CONDITIONAL") == "from_env"
    else:
        assert "CONDITIONAL" not in set_env


def test_set_env_cross_section_override(tox_project: ToxProjectCreator) -> None:
    ini = """\
    [testenv]
    skip_install = true
    set_env =
        OS_TEST_PATH=./tests/unit
    commands = python -c "import os; print(os.environ['OS_TEST_PATH'])"

    [testenv:functional]
    set_env =
      {[testenv]set_env}
      OS_TEST_PATH=./tests/functional
    commands = python -c "import os; print(os.environ['OS_TEST_PATH'])"

    [testenv:functional-py]
    set_env = {[testenv:functional]set_env}
    commands = {[testenv:functional]commands}
    """
    result = tox_project({"tox.ini": ini}).run("r", "-e", "functional-py")
    result.assert_success()
    assert result.out.splitlines()[1] == "./tests/functional"


def test_set_env_cross_section_override_direct(tox_project: ToxProjectCreator) -> None:
    ini = """\
    [testenv]
    skip_install = true
    set_env =
        COVERAGE_FILE=THISISBAD

    [testenv:coverage_report]
    set_env =
        {[testenv]set_env}
        COVERAGE_FILE=THISISGOOD
    commands = python -c "import os; print(os.environ['COVERAGE_FILE'])"
    """
    result = tox_project({"tox.ini": ini}).run("r", "-e", "coverage_report")
    result.assert_success()
    assert result.out.splitlines()[1] == "THISISGOOD"


def test_set_env_marker_mixed(eval_set_env: EvalSetEnv) -> None:
    marker = f"sys_platform == '{sys.platform}'"
    config = (
        f"[testenv]\npackage=skip\nset_env=ALWAYS=1\n CONDITIONAL=2; {marker}\n NEVER=3; sys_platform == 'nonexistent'"
    )
    set_env = eval_set_env(config)
    keys = list(set_env)
    assert "ALWAYS" in keys
    assert "CONDITIONAL" in keys
    assert "NEVER" not in keys



from __future__ import annotations

import re
from typing import TYPE_CHECKING

import tomli_w
from colorama import Fore

from .common import build_structured_result, write_output

if TYPE_CHECKING:
    from tox.session.state import State

_HEADER_RE = re.compile(
    r"""
    ^ \[ .* ] $     # section header like [env.py]
    """,
    re.MULTILINE | re.VERBOSE,
)
_KEY_RE = re.compile(
    r"""
    ^ ( [a-zA-Z_]       # key starts with letter or underscore
        [a-zA-Z0-9_-]*  # followed by alphanumerics, underscores, or hyphens
      )
    \s* =               # equals sign with optional whitespace
    """,
    re.MULTILINE | re.VERBOSE,
)


def show_config_toml(state: State) -> int:
    result, has_exception = build_structured_result(state)
    output = tomli_w.dumps(result).removesuffix("\n")
    write_output(output, state.conf.options.output_file, state.conf.options.is_colored, colorize)
    return -1 if has_exception else 0


def colorize(text: str) -> str:
    text = _HEADER_RE.sub(lambda m: f"{Fore.YELLOW}{m.group()}{Fore.RESET}", text)
    return _KEY_RE.sub(rf"{Fore.GREEN}\1{Fore.RESET} =", text)



from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING

from colorama import Fore

from .common import build_structured_result, write_output

if TYPE_CHECKING:
    from tox.session.state import State

_KEY_RE = re.compile(
    r"""
    ^ ([ ]*)        # leading indentation
    " ([^"]+) "     # quoted key name
    :               # colon separator
    """,
    re.MULTILINE | re.VERBOSE,
)


def show_config_json(state: State) -> int:
    result, has_exception = build_structured_result(state)
    output = json.dumps(result, indent=2)
    write_output(output, state.conf.options.output_file, state.conf.options.is_colored, colorize)
    return -1 if has_exception else 0


def colorize(text: str) -> str:
    return _KEY_RE.sub(rf"\1{Fore.GREEN}\2{Fore.RESET}:", text)



from __future__ import annotations

import os

_ENV_VARS = {  # per https://adamj.eu/tech/2020/03/09/detect-if-your-tests-are-running-on-ci
    "CI": None,  # generic flag
    "TF_BUILD": "true",  # Azure Pipelines
    "bamboo.buildKey": None,  # Bamboo
    "BUILDKITE": "true",  # Buildkite
    "CIRCLECI": "true",  # Circle CI
    "CIRRUS_CI": "true",  # Cirrus CI
    "CODEBUILD_BUILD_ID": None,  # CodeBuild
    "GITHUB_ACTIONS": "true",  # GitHub Actions
    "GITLAB_CI": None,  # GitLab CI
    "HEROKU_TEST_RUN_ID": None,  # Heroku CI
    "BUILD_ID": None,  # Hudson
    "TEAMCITY_VERSION": None,  # TeamCity
    "TRAVIS": "true",  # Travis CI
}


def is_ci() -> bool:
    """:returns: a flag indicating if running inside a CI env or not"""
    for env_key, value in _ENV_VARS.items():
        if env_key in os.environ if value is None else os.environ.get(env_key) == value:
            if env_key == "TEAMCITY_VERSION" and os.environ.get(env_key) == "LOCAL":
                continue
            return True
    return False


__all__ = [
    "is_ci",
]
